from __future__ import annotations

import asyncio
import base64
import contextlib
import contextvars
import hashlib
import ipaddress
import json
import logging
import os
import socket
import time
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Awaitable,
    Callable,
    Mapping,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
)
from urllib.parse import ParseResult, urlparse, urlunparse

import aiohttp
from aiohttp import hdrs
from aiohttp.client_ws import DEFAULT_WS_CLIENT_TIMEOUT
from aiohttp.http_websocket import WS_KEY, WebSocketReader, WebSocketWriter
from multidict import CIMultiDict

from app.core.clients.http import get_http_client
from app.core.config.settings import get_settings
from app.core.errors import (
    OpenAIErrorDetail,
    OpenAIErrorEnvelope,
    ResponseFailedEvent,
    openai_error,
    response_failed_event,
)
from app.core.openai.model_registry import get_model_registry
from app.core.openai.models import CompactResponsePayload, OpenAIError
from app.core.openai.parsing import (
    parse_compact_response_payload,
    parse_error_payload,
    parse_sse_event,
)
from app.core.openai.requests import ResponsesCompactRequest, ResponsesRequest
from app.core.resilience.circuit_breaker import (
    CircuitBreakerOpenError,
    _is_server_error,
    get_circuit_breaker_for_account,
)
from app.core.types import JsonObject, JsonValue
from app.core.utils.request_id import get_request_id
from app.core.utils.sse import format_sse_event

IGNORE_INBOUND_HEADERS = {
    "authorization",
    "chatgpt-account-id",
    "content-length",
    "host",
    "forwarded",
    "x-real-ip",
    "true-client-ip",
}

_ERROR_TYPE_CODE_MAP = {
    "rate_limit_exceeded": "rate_limit_exceeded",
    "usage_not_included": "usage_not_included",
    "insufficient_quota": "insufficient_quota",
    "quota_exceeded": "quota_exceeded",
}

_SSE_EVENT_TYPE_ALIASES = {
    "response.text.delta": "response.output_text.delta",
    "response.audio.delta": "response.output_audio.delta",
    "response.audio_transcript.delta": "response.output_audio_transcript.delta",
}

_SSE_READ_CHUNK_SIZE = 1 * 1024
_IMAGE_INLINE_MAX_BYTES = 8 * 1024 * 1024
_IMAGE_INLINE_CHUNK_SIZE = 64 * 1024
_IMAGE_INLINE_TIMEOUT_SECONDS = 8.0
_BLOCKED_LITERAL_HOSTS = {"localhost", "localhost.localdomain"}
_UPSTREAM_TRACE_HEADER_ALLOWLIST = frozenset(
    {
        "accept",
        "chatgpt-account-id",
        "content-type",
        "request-id",
        "session_id",
        "user-agent",
        "x-codex-conversation-id",
        "x-codex-session-id",
        "x-openai-client-arch",
        "x-openai-client-id",
        "x-openai-client-os",
        "x-openai-client-user-agent",
        "x-openai-client-version",
        "x-request-id",
    }
)
_NATIVE_CODEX_ORIGINATORS = frozenset(
    {
        "Codex Desktop",
        "codex_atlas",
        "codex_chatgpt_desktop",
        "codex_cli_rs",
        "codex_exec",
        "codex_vscode",
    }
)
_NATIVE_CODEX_STREAM_HEADER_KEYS = frozenset(
    {
        "x-codex-turn-state",
        "x-codex-turn-metadata",
        "x-codex-beta-features",
    }
)


def _normalize_chatgpt_web_service_tier(value: object) -> str | None:
    if not isinstance(value, str):
        return None
    stripped = value.strip()
    if not stripped:
        return None
    if stripped.lower() == "fast":
        return "priority"
    return stripped


def _normalize_chatgpt_web_payload(payload_dict: JsonObject) -> JsonObject:
    normalized = dict(payload_dict)
    service_tier = _normalize_chatgpt_web_service_tier(normalized.get("service_tier"))
    if service_tier is not None:
        normalized["service_tier"] = service_tier
    return normalized


_HOP_BY_HOP_HEADER_NAMES = frozenset(
    {
        "accept",
        "connection",
        "content-type",
        "keep-alive",
        "proxy-authenticate",
        "proxy-authorization",
        "proxy-connection",
        "te",
        "trailer",
        "transfer-encoding",
        "upgrade",
    }
)
_AUTO_WEBSOCKET_HANDSHAKE_FALLBACK_STATUSES = frozenset({426})
_WEBSOCKET_RESPONSE_CREATE_EXCLUDED_FIELDS = frozenset({"background", "stream"})
_WEBSOCKET_HANDSHAKE_ERROR_HINTS = (
    ("usage_not_included", "usage not included"),
    ("insufficient_quota", "insufficient quota"),
    ("quota_exceeded", "quota exceeded"),
    ("usage_limit_reached", "usage limit reached"),
    ("rate_limit_exceeded", "rate limit"),
)

logger = logging.getLogger(__name__)
_STREAM_CONNECT_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "stream_connect_timeout_override",
    default=None,
)
_STREAM_IDLE_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "stream_idle_timeout_override",
    default=None,
)
_STREAM_TOTAL_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "stream_total_timeout_override",
    default=None,
)
_COMPACT_CONNECT_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "compact_connect_timeout_override",
    default=None,
)
_COMPACT_TOTAL_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "compact_total_timeout_override",
    default=None,
)
_TRANSCRIBE_CONNECT_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "transcribe_connect_timeout_override",
    default=None,
)
_TRANSCRIBE_TOTAL_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "transcribe_total_timeout_override",
    default=None,
)

R = TypeVar("R")


async def _call_with_service_circuit_breaker(
    request: Awaitable[R],
    *,
    settings: object | None = None,
    account_id: str | None = None,
) -> R:
    if not account_id:
        return await request
    effective_settings = settings or get_settings()
    circuit_breaker = get_circuit_breaker_for_account(account_id, effective_settings)
    if circuit_breaker is None:
        return await request
    return await circuit_breaker.call(request)


@asynccontextmanager
async def _service_circuit_breaker_context(
    cm: AsyncContextManager[aiohttp.ClientResponse],
    *,
    settings: object | None = None,
    account_id: str | None = None,
) -> AsyncIterator[aiohttp.ClientResponse]:
    """Wrap an async context manager with circuit breaker protection."""
    effective_settings = settings or get_settings()
    cb = get_circuit_breaker_for_account(account_id, effective_settings) if account_id else None
    is_probe = False
    if cb is not None:
        try:
            is_probe = await cb.pre_call_check()
        except BaseException:
            close = getattr(cm, "close", None)
            if callable(close):
                close()
            raise
    resp_ref: aiohttp.ClientResponse | None = None
    try:
        async with cm as resp:
            resp_ref = resp
            yield resp
        if cb is not None:
            if hasattr(resp, "status") and resp.status >= 500:
                await cb._record_failure(Exception(f"HTTP {resp.status}"))
            else:
                await cb._record_success()
    except CircuitBreakerOpenError:
        raise
    except Exception as e:
        if cb is not None:
            if (
                resp_ref is not None
                and hasattr(resp_ref, "status")
                and resp_ref.status < 500
                and not _is_server_error(e)
            ):
                await cb._record_success()
            else:
                await cb._record_failure(e)
        raise
    finally:
        if is_probe and cb is not None:
            await cb.release_half_open_probe()


_HELD_HALF_OPEN_PROBE_FLAG = "_codex_lb_half_open_probe_held"
_HELD_HALF_OPEN_PROBE_BREAKER = "_codex_lb_half_open_probe_breaker"


def _bind_half_open_probe(websocket: aiohttp.ClientWebSocketResponse, circuit_breaker: object) -> None:
    setattr(websocket, _HELD_HALF_OPEN_PROBE_FLAG, True)
    setattr(websocket, _HELD_HALF_OPEN_PROBE_BREAKER, circuit_breaker)


async def _release_bound_half_open_probe(websocket: aiohttp.ClientWebSocketResponse | None) -> None:
    if websocket is None or not getattr(websocket, _HELD_HALF_OPEN_PROBE_FLAG, False):
        return
    circuit_breaker = getattr(websocket, _HELD_HALF_OPEN_PROBE_BREAKER, None)
    setattr(websocket, _HELD_HALF_OPEN_PROBE_FLAG, False)
    setattr(websocket, _HELD_HALF_OPEN_PROBE_BREAKER, None)
    if circuit_breaker is not None:
        await cast(Any, circuit_breaker).release_half_open_probe()


class StreamIdleTimeoutError(Exception):
    pass


class StreamEventTooLargeError(Exception):
    def __init__(self, size_bytes: int, limit_bytes: int) -> None:
        super().__init__(f"SSE event exceeded {limit_bytes} bytes (received {size_bytes} bytes)")
        self.size_bytes = size_bytes
        self.limit_bytes = limit_bytes


class ErrorResponseProtocol(Protocol):
    status: int
    reason: str | None

    async def json(self, *, content_type: str | None = None) -> JsonValue: ...

    async def text(self, *, encoding: str | None = None, errors: str = "strict") -> str: ...


ErrorResponse: TypeAlias = aiohttp.ClientResponse | ErrorResponseProtocol


class SSEContentProtocol(Protocol):
    def iter_chunked(self, size: int) -> "SSEChunkIteratorProtocol": ...


class SSEChunkIteratorProtocol(Protocol):
    def __aiter__(self) -> "SSEChunkIteratorProtocol": ...

    def __anext__(self) -> Awaitable[bytes]: ...


class SSEResponseProtocol(Protocol):
    content: SSEContentProtocol


SSEResponse: TypeAlias = aiohttp.ClientResponse | SSEResponseProtocol


class ProxyResponseError(Exception):
    status_code: int
    payload: OpenAIErrorEnvelope

    def __init__(
        self,
        status_code: int,
        payload: OpenAIErrorEnvelope,
        *,
        failure_phase: str | None = None,
        retryable_same_contract: bool = False,
        failure_detail: str | None = None,
        failure_exception_type: str | None = None,
        upstream_status_code: int | None = None,
    ) -> None:
        super().__init__(f"Proxy response error ({status_code})")
        self.status_code = status_code
        self.payload = payload
        self.failure_phase = failure_phase
        self.retryable_same_contract = retryable_same_contract
        self.failure_detail = failure_detail
        self.failure_exception_type = failure_exception_type
        self.upstream_status_code = upstream_status_code


def _should_drop_inbound_header(name: str) -> bool:
    normalized = name.lower()
    if normalized in IGNORE_INBOUND_HEADERS:
        return True
    if normalized.startswith("x-forwarded-"):
        return True
    if normalized.startswith("cf-"):
        return True
    return False


def filter_inbound_headers(headers: Mapping[str, str]) -> dict[str, str]:
    return {key: value for key, value in headers.items() if not _should_drop_inbound_header(key)}


def _build_upstream_headers(
    inbound: Mapping[str, str],
    access_token: str,
    account_id: str | None,
    accept: str = "text/event-stream",
) -> dict[str, str]:
    headers = dict(inbound)
    lower_keys = {key.lower() for key in headers}
    if "x-request-id" not in lower_keys and "request-id" not in lower_keys:
        request_id = get_request_id()
        if request_id:
            headers["x-request-id"] = request_id
    headers["Authorization"] = f"Bearer {access_token}"
    headers["Accept"] = accept
    headers["Content-Type"] = "application/json"
    if account_id:
        headers["chatgpt-account-id"] = account_id
    return headers


_TRANSCRIBE_FORWARD_HEADER_PREFIXES = ("x-openai-", "x-codex-")


def _build_upstream_transcribe_headers(
    inbound: Mapping[str, str],
    access_token: str,
    account_id: str | None,
) -> dict[str, str]:
    # Minimal header set matching Codex CLI ``/transcribe`` fingerprint.
    # Omit Accept, x-request-id, and bulk-forwarded inbound headers to
    # avoid upstream WAF rejection.
    headers: dict[str, str] = {}
    headers["Authorization"] = f"Bearer {access_token}"
    if account_id:
        headers["chatgpt-account-id"] = account_id
    for key, value in inbound.items():
        lower = key.lower()
        if lower == "user-agent":
            headers[key] = value
        elif lower.startswith(_TRANSCRIBE_FORWARD_HEADER_PREFIXES):
            headers[key] = value
    return headers


def _build_upstream_websocket_headers(
    inbound: Mapping[str, str],
    access_token: str,
    account_id: str | None,
) -> dict[str, str]:
    connected_header_tokens: set[str] = set()
    for key, value in inbound.items():
        if key.lower() != "connection":
            continue
        connected_header_tokens.update(
            token.strip().lower() for token in value.split(",") if isinstance(value, str) and token.strip()
        )
    blocked_header_names = _HOP_BY_HOP_HEADER_NAMES | connected_header_tokens
    headers = {key: value for key, value in inbound.items() if key.lower() not in blocked_header_names}
    lower_keys = {key.lower() for key in headers}
    if "x-request-id" not in lower_keys and "request-id" not in lower_keys:
        request_id = get_request_id()
        if request_id:
            headers["x-request-id"] = request_id
    headers["Authorization"] = f"Bearer {access_token}"
    if account_id:
        headers["chatgpt-account-id"] = account_id
    return headers


def _interesting_upstream_header_keys(headers: Mapping[str, str]) -> list[str]:
    return sorted({key.lower() for key in headers if key.lower() in _UPSTREAM_TRACE_HEADER_ALLOWLIST})


def _summarize_upstream_target(url: str) -> str:
    parsed = urlparse(url)
    return urlunparse((parsed.scheme, parsed.netloc, parsed.path, "", "", ""))


def _summarize_input_value(value: JsonValue | None) -> str:
    if value is None:
        return "0"
    if isinstance(value, str):
        return "str"
    if isinstance(value, list):
        if not value:
            return "0"
        type_counts: dict[str, int] = {}
        for item in value:
            type_name = type(item).__name__
            type_counts[type_name] = type_counts.get(type_name, 0) + 1
        summary = ",".join(f"{key}={type_counts[key]}" for key in sorted(type_counts))
        return f"{len(value)}({summary})"
    return type(value).__name__


def _summarize_json_payload(payload: Mapping[str, JsonValue]) -> str:
    keys = sorted(payload.keys())
    model = payload.get("model")
    stream = payload.get("stream")
    input_summary = _summarize_input_value(payload.get("input"))
    return f"model={model} stream={stream} input={input_summary} keys={keys}"


def _summarize_transcription_payload(
    *,
    filename: str,
    content_type: str | None,
    prompt: str | None,
    audio_bytes: bytes,
) -> dict[str, JsonValue]:
    return {
        "filename": filename,
        "content_type": content_type,
        "prompt_present": prompt is not None,
        "audio_bytes": len(audio_bytes),
    }


def _error_details_from_envelope(payload: OpenAIErrorEnvelope) -> tuple[str | None, str | None]:
    error = payload.get("error")
    if not isinstance(error, dict):
        return None, None
    code = error.get("code")
    message = error.get("message")
    return code if isinstance(code, str) else None, message if isinstance(message, str) else None


def _error_details_from_failed_event(payload: ResponseFailedEvent) -> tuple[str | None, str | None]:
    response = payload.get("response")
    if not isinstance(response, dict):
        return None, None
    error = response.get("error")
    if not isinstance(error, dict):
        return None, None
    code = error.get("code")
    message = error.get("message")
    return code if isinstance(code, str) else None, message if isinstance(message, str) else None


def _extract_json_object_from_text(text: str) -> JsonValue | None:
    if not text:
        return None
    decoder = json.JSONDecoder()
    for index, char in enumerate(text):
        if char != "{":
            continue
        try:
            value, _ = decoder.raw_decode(text[index:])
        except json.JSONDecodeError:
            continue
        return value
    return None


def _infer_websocket_handshake_error_code(status: int | None, message: str) -> str:
    lowered = message.lower()
    for code, hint in _WEBSOCKET_HANDSHAKE_ERROR_HINTS:
        if hint in lowered:
            return code
    if status == 401:
        return "invalid_api_key"
    if status == 404:
        return "not_found"
    if status == 429:
        return "rate_limit_exceeded"
    return "upstream_error"


def _error_payload_from_websocket_handshake_error(exc: aiohttp.WSServerHandshakeError) -> OpenAIErrorEnvelope:
    message = exc.message or str(exc)
    extracted = _extract_json_object_from_text(message)
    if extracted is not None:
        error = parse_error_payload(extracted)
        if error is not None:
            return {"error": _openai_error_detail(error)}

    code = _infer_websocket_handshake_error_code(exc.status, message)
    if code == "invalid_api_key":
        return openai_error(code, message, error_type="authentication_error")
    if code == "not_found":
        return openai_error(code, message, error_type="invalid_request_error")
    if code == "rate_limit_exceeded":
        return openai_error(code, message, error_type="rate_limit_error")
    return openai_error(code, message)


def _maybe_log_upstream_request_start(
    *,
    kind: str,
    url: str,
    headers: Mapping[str, str],
    method: str,
    payload_summary: str,
    payload_json: str | None = None,
) -> None:
    settings = get_settings()
    if not settings.log_upstream_request_summary and not settings.log_upstream_request_payload:
        return

    request_id = get_request_id()
    target = _summarize_upstream_target(url)
    account_id = headers.get("chatgpt-account-id")
    route_class = _route_class_for_upstream_target(url)
    header_keys = _interesting_upstream_header_keys(headers)

    if settings.log_upstream_request_summary:
        logger.info(
            (
                "upstream_request_start request_id=%s kind=%s method=%s target=%s "
                "provider_kind=%s route_class=%s routing_subject_id=%s account_id=%s "
                "upstream_request_id=%s headers=%s payload=%s"
            ),
            request_id,
            kind,
            method,
            target,
            "chatgpt_web",
            route_class,
            account_id,
            account_id,
            None,
            header_keys,
            payload_summary,
        )
    if settings.log_upstream_request_payload and payload_json is not None:
        logger.info(
            "upstream_request_payload request_id=%s kind=%s target=%s payload=%s",
            request_id,
            kind,
            target,
            payload_json,
        )


def _maybe_log_upstream_request_complete(
    *,
    kind: str,
    url: str,
    headers: Mapping[str, str],
    method: str,
    started_at: float,
    status_code: int | None,
    error_code: str | None,
    error_message: str | None,
    failure_phase: str | None = None,
    payload_object: str | None = None,
    failure_detail: str | None = None,
    failure_exception_type: str | None = None,
    retryable_same_contract: bool | None = None,
    upstream_request_id: str | None = None,
) -> None:
    settings = get_settings()
    if not settings.log_upstream_request_summary:
        return

    level = logging.INFO
    if status_code is not None and status_code >= 500:
        level = logging.ERROR
    elif (status_code is not None and status_code >= 400) or error_code is not None:
        level = logging.WARNING

    logger.log(
        level,
        (
            "upstream_request_complete request_id=%s kind=%s method=%s target=%s "
            "provider_kind=%s route_class=%s routing_subject_id=%s account_id=%s "
            "status=%s duration_ms=%s error_code=%s error_message=%s "
            "failure_phase=%s payload_object=%s failure_detail=%s failure_exception_type=%s "
            "retryable_same_contract=%s upstream_request_id=%s"
        ),
        get_request_id(),
        kind,
        method,
        _summarize_upstream_target(url),
        "chatgpt_web",
        _route_class_for_upstream_target(url),
        headers.get("chatgpt-account-id"),
        headers.get("chatgpt-account-id"),
        status_code,
        int((time.monotonic() - started_at) * 1000),
        error_code,
        error_message,
        failure_phase,
        payload_object,
        failure_detail,
        failure_exception_type,
        retryable_same_contract,
        upstream_request_id,
    )


def _route_class_for_upstream_target(url: str) -> str:
    if "/v1/" in url:
        return "openai_public_http"
    return "chatgpt_private"


def _normalize_error_code(code: str | None, error_type: str | None) -> str:
    if code:
        normalized_code = code.lower()
        mapped = _ERROR_TYPE_CODE_MAP.get(normalized_code)
        return mapped or normalized_code
    normalized_type = error_type.lower() if error_type else None
    if normalized_type:
        mapped = _ERROR_TYPE_CODE_MAP.get(normalized_type)
        return mapped or normalized_type
    return "upstream_error"


def _effective_stream_timeout(configured_timeout_seconds: float, timeout_kind: str) -> float:
    if timeout_kind == "connect":
        override = _STREAM_CONNECT_TIMEOUT_OVERRIDE.get()
    elif timeout_kind == "idle":
        override = _STREAM_IDLE_TIMEOUT_OVERRIDE.get()
    else:
        override = _STREAM_TOTAL_TIMEOUT_OVERRIDE.get()
    if override is None:
        return configured_timeout_seconds
    return max(0.001, min(configured_timeout_seconds, override))


def _effective_compact_connect_timeout(configured_timeout_seconds: float) -> float:
    override = _COMPACT_CONNECT_TIMEOUT_OVERRIDE.get()
    if override is None:
        return configured_timeout_seconds
    return max(0.001, min(configured_timeout_seconds, override))


def _effective_compact_total_timeout(configured_timeout_seconds: float | None) -> float | None:
    override = _COMPACT_TOTAL_TIMEOUT_OVERRIDE.get()
    if configured_timeout_seconds is None:
        return None if override is None else max(0.001, override)
    if override is None:
        return configured_timeout_seconds
    return max(0.001, min(configured_timeout_seconds, override))


def _effective_transcribe_connect_timeout(configured_timeout_seconds: float) -> float:
    override = _TRANSCRIBE_CONNECT_TIMEOUT_OVERRIDE.get()
    if override is None:
        return configured_timeout_seconds
    return max(0.001, min(configured_timeout_seconds, override))


def _effective_transcribe_total_timeout(configured_timeout_seconds: float) -> float:
    override = _TRANSCRIBE_TOTAL_TIMEOUT_OVERRIDE.get()
    if override is None:
        return configured_timeout_seconds
    return max(0.001, min(configured_timeout_seconds, override))


def _remaining_total_timeout(timeout_seconds: float | None, started_at: float, now: float) -> float | None:
    if timeout_seconds is None:
        return None
    return max(0.001, timeout_seconds - max(0.0, now - started_at))


def _find_sse_separator(buffer: bytes | bytearray) -> tuple[int, int] | None:
    separators = (b"\r\n\r\n", b"\n\n")
    positions = [(buffer.find(separator), len(separator)) for separator in separators]
    valid_positions = [position for position in positions if position[0] >= 0]
    if not valid_positions:
        return None
    return min(valid_positions, key=lambda item: item[0])


def _pop_sse_event(buffer: bytearray) -> bytes | None:
    separator = _find_sse_separator(buffer)
    if separator is None:
        return None
    index, separator_len = separator
    event_end = index + separator_len
    event = bytes(buffer[:event_end])
    del buffer[:event_end]
    return event


async def _iter_sse_events(
    resp: SSEResponse,
    idle_timeout_seconds: float,
    max_event_bytes: int,
) -> AsyncIterator[str]:
    async def _next_chunk() -> bytes:
        return await iterator.__anext__()

    async def _cancel_pending_chunk(task: asyncio.Task[bytes]) -> None:
        if task.done():
            return
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

    buffer = bytearray()
    chunk_iterator = resp.content.iter_chunked(_SSE_READ_CHUNK_SIZE)
    iterator = chunk_iterator.__aiter__()

    while True:
        next_chunk = asyncio.create_task(_next_chunk())
        try:
            done, _ = await asyncio.wait({next_chunk}, timeout=idle_timeout_seconds)
            if not done:
                await _cancel_pending_chunk(next_chunk)
                raise StreamIdleTimeoutError()
            chunk = await next_chunk
        except StopAsyncIteration:
            break
        except asyncio.CancelledError:
            await _cancel_pending_chunk(next_chunk)
            raise

        if not chunk:
            continue

        buffer.extend(chunk)
        while True:
            raw_event = _pop_sse_event(buffer)
            if raw_event is None:
                if len(buffer) > max_event_bytes:
                    raise StreamEventTooLargeError(len(buffer), max_event_bytes)
                break

            if len(raw_event) > max_event_bytes:
                raise StreamEventTooLargeError(len(raw_event), max_event_bytes)

            if raw_event.strip():
                yield raw_event.decode("utf-8", errors="replace")

    if buffer:
        if len(buffer) > max_event_bytes:
            raise StreamEventTooLargeError(len(buffer), max_event_bytes)
        yield bytes(buffer).decode("utf-8", errors="replace")


async def _error_event_from_response(resp: ErrorResponse) -> ResponseFailedEvent:
    fallback_message = f"Upstream error: HTTP {resp.status}"
    if resp.reason:
        fallback_message += f" {resp.reason}"
    try:
        data = await resp.json(content_type=None)
    except Exception:
        text = await resp.text()
        message = text.strip() or fallback_message
        return response_failed_event("upstream_error", message, response_id=get_request_id())

    if isinstance(data, dict):
        error = parse_error_payload(data)
        if error:
            payload = error.model_dump(exclude_none=True)
            event = response_failed_event(
                _normalize_error_code(payload.get("code"), payload.get("type")),
                payload.get("message", fallback_message),
                error_type=payload.get("type") or "server_error",
                response_id=get_request_id(),
                error_param=payload.get("param"),
            )
            for key in ("plan_type", "resets_at", "resets_in_seconds"):
                if key in payload:
                    event["response"]["error"][key] = payload[key]
            return event
        message = _extract_upstream_message(data)
        if message:
            return response_failed_event("upstream_error", message, response_id=get_request_id())
    return response_failed_event("upstream_error", fallback_message, response_id=get_request_id())


async def _error_payload_from_response(resp: ErrorResponse) -> OpenAIErrorEnvelope:
    fallback_message = f"Upstream error: HTTP {resp.status}"
    if resp.reason:
        fallback_message += f" {resp.reason}"
    try:
        data = await resp.json(content_type=None)
    except Exception:
        text = await resp.text()
        message = text.strip() or fallback_message
        return openai_error("upstream_error", message)

    if isinstance(data, dict):
        error = parse_error_payload(data)
        if error:
            return {"error": _openai_error_detail(error)}
        message = _extract_upstream_message(data)
        if message:
            return openai_error("upstream_error", message)
    return openai_error("upstream_error", fallback_message)


def _openai_error_detail(error: OpenAIError) -> OpenAIErrorDetail:
    detail: OpenAIErrorDetail = {}
    if error.message is not None:
        detail["message"] = error.message
    if error.type is not None:
        detail["type"] = error.type
    if error.code is not None:
        detail["code"] = error.code
    if error.param is not None:
        detail["param"] = error.param
    if error.plan_type is not None:
        detail["plan_type"] = error.plan_type
    if error.resets_at is not None:
        detail["resets_at"] = error.resets_at
    if error.resets_in_seconds is not None:
        detail["resets_in_seconds"] = error.resets_in_seconds
    return detail


def _extract_upstream_message(data: Mapping[str, object]) -> str | None:
    for key in ("message", "detail", "error"):
        value = data.get(key)
        if isinstance(value, str) and value.strip():
            return value
    return None


def _normalize_sse_data_line(line: str) -> str:
    if not line.startswith("data:"):
        return line
    data = line[5:].strip()
    if not data or data == "[DONE]":
        return line
    try:
        payload = json.loads(data)
    except json.JSONDecodeError:
        return line
    if not isinstance(payload, dict):
        return line
    event_type = payload.get("type")
    if isinstance(event_type, str) and event_type in _SSE_EVENT_TYPE_ALIASES:
        payload["type"] = _SSE_EVENT_TYPE_ALIASES[event_type]
        return f"data: {json.dumps(payload, ensure_ascii=True, separators=(',', ':'))}"
    return line


def _normalize_sse_event_block(event_block: str) -> str:
    if not event_block:
        return event_block

    if '"type":' not in event_block:
        return event_block

    if event_block.endswith("\r\n\r\n"):
        line_separator = "\r\n"
        terminator = "\r\n\r\n"
        body = event_block[: -len(terminator)]
    elif event_block.endswith("\n\n"):
        line_separator = "\n"
        terminator = "\n\n"
        body = event_block[: -len(terminator)]
    else:
        line_separator = "\r\n" if "\r\n" in event_block else "\n"
        terminator = ""
        body = event_block

    lines = body.splitlines()
    if not lines:
        return event_block

    normalized_lines: list[str] = []
    changed = False
    for line in lines:
        normalized_line = _normalize_sse_data_line(line)
        if normalized_line != line:
            changed = True
        normalized_lines.append(normalized_line)
    if not changed:
        return event_block

    normalized = line_separator.join(normalized_lines)
    if terminator:
        return normalized + terminator
    return normalized


def _normalize_stream_event_payload(payload: dict[str, JsonValue]) -> dict[str, JsonValue]:
    event_type = payload.get("type")
    if isinstance(event_type, str) and event_type in _SSE_EVENT_TYPE_ALIASES:
        normalized = dict(payload)
        normalized["type"] = _SSE_EVENT_TYPE_ALIASES[event_type]
        return normalized
    return payload


def _to_websocket_upstream_url(url: str) -> str:
    parsed = urlparse(url)
    if parsed.scheme == "https":
        scheme = "wss"
    elif parsed.scheme == "http":
        scheme = "ws"
    else:
        scheme = parsed.scheme
    return urlunparse((scheme, parsed.netloc, parsed.path, parsed.params, parsed.query, parsed.fragment))


def _configured_stream_transport(
    *,
    transport: str,
    transport_override: str | None = None,
) -> str:
    return transport_override if transport_override is not None else transport


def _has_native_codex_transport_headers(headers: Mapping[str, str]) -> bool:
    normalized = {key.lower(): value for key, value in headers.items()}
    originator = normalized.get("originator")
    if _is_native_codex_originator(originator):
        return True
    return any(key in normalized for key in _NATIVE_CODEX_STREAM_HEADER_KEYS)


def _is_native_codex_originator(originator: object) -> bool:
    if not isinstance(originator, str):
        return False
    stripped = originator.strip()
    if not stripped:
        return False
    return stripped in _NATIVE_CODEX_ORIGINATORS


def _resolve_stream_transport(
    *,
    transport: str,
    transport_override: str | None,
    model: str | None,
    headers: Mapping[str, str],
) -> str:
    configured = _configured_stream_transport(transport=transport, transport_override=transport_override)
    if configured == "websocket":
        return "websocket"
    if configured == "http":
        return "http"
    if _has_native_codex_transport_headers(headers):
        return "websocket"

    registry = get_model_registry()
    prefers_websockets = getattr(registry, "prefers_websockets", None)
    if callable(prefers_websockets):
        if prefers_websockets(model):
            return "websocket"
        return "http"

    snapshot = registry.get_snapshot()
    if snapshot is None or not isinstance(model, str):
        return "http"
    upstream_model = snapshot.models.get(model)
    if upstream_model and upstream_model.prefer_websockets:
        return "websocket"
    return "http"


def _should_fallback_to_http_after_websocket_handshake_error(
    transport_mode: str,
    exc: aiohttp.WSServerHandshakeError,
) -> bool:
    return transport_mode == "auto" and exc.status in _AUTO_WEBSOCKET_HANDSHAKE_FALLBACK_STATUSES


async def _open_upstream_websocket(
    *,
    session: aiohttp.ClientSession,
    url: str,
    headers: Mapping[str, str],
    connect_timeout_seconds: float,
    max_msg_size: int,
    account_id: str | None = None,
    hold_half_open_probe: bool = False,
) -> tuple[AsyncContextManager[aiohttp.ClientWebSocketResponse], aiohttp.ClientWebSocketResponse]:
    settings = get_settings()
    circuit_breaker = get_circuit_breaker_for_account(account_id, settings) if account_id else None
    is_probe = False
    if circuit_breaker is not None:
        is_probe = await circuit_breaker.pre_call_check()

    request_obj = getattr(session, "request", None)
    if not callable(request_obj):
        try:
            websocket_cm = session.ws_connect(
                url,
                headers=headers,
                receive_timeout=None,
                autoping=True,
                autoclose=True,
                max_msg_size=max_msg_size,
            )
            websocket = await asyncio.wait_for(websocket_cm.__aenter__(), timeout=connect_timeout_seconds)
            if hold_half_open_probe and is_probe and circuit_breaker is not None:
                _bind_half_open_probe(websocket, circuit_breaker)
            return websocket_cm, websocket
        except Exception as exc:
            if circuit_breaker is not None:
                await circuit_breaker._record_failure(exc)
            raise
        finally:
            if is_probe and circuit_breaker is not None and not hold_half_open_probe:
                await circuit_breaker.release_half_open_probe()
    request = cast(Callable[..., Awaitable[aiohttp.ClientResponse]], request_obj)

    request_headers = CIMultiDict(headers)
    request_headers.setdefault(hdrs.UPGRADE, "websocket")
    request_headers.setdefault(hdrs.CONNECTION, "Upgrade")
    request_headers.setdefault(hdrs.SEC_WEBSOCKET_VERSION, "13")
    sec_key = base64.b64encode(os.urandom(16)).decode()
    request_headers[hdrs.SEC_WEBSOCKET_KEY] = sec_key

    timeout = aiohttp.ClientTimeout(total=connect_timeout_seconds, sock_connect=connect_timeout_seconds)
    request_fn = cast(Any, request)
    try:
        try:
            resp = await request_fn(
                hdrs.METH_GET,
                url,
                headers=request_headers,
                timeout=timeout,
                read_until_eof=False,
            )
        except Exception as exc:
            if circuit_breaker is not None:
                await circuit_breaker._record_failure(exc)
            raise

        async def _raise_handshake_error(message: str) -> None:
            body_text = ""
            try:
                body_text = (await resp.text()).strip()
            except Exception:
                body_text = ""
            raise aiohttp.WSServerHandshakeError(
                resp.request_info,
                resp.history,
                message=body_text or message,
                status=resp.status,
                headers=resp.headers,
            )

        _cb_recorded = False
        try:
            if circuit_breaker is not None:
                if resp.status >= 500:
                    await circuit_breaker._record_failure(Exception(f"WebSocket handshake failed: HTTP {resp.status}"))
                    _cb_recorded = True
                elif resp.status != 101:
                    await circuit_breaker._record_success()
                    _cb_recorded = True

            if resp.status != 101:
                await _raise_handshake_error("Invalid response status")

            if resp.headers.get(hdrs.UPGRADE, "").lower() != "websocket":
                await _raise_handshake_error("Invalid upgrade header")

            if resp.headers.get(hdrs.CONNECTION, "").lower() != "upgrade":
                await _raise_handshake_error("Invalid connection header")

            response_key = resp.headers.get(hdrs.SEC_WEBSOCKET_ACCEPT, "")
            expected_key = base64.b64encode(hashlib.sha1(sec_key.encode() + WS_KEY).digest()).decode()
            if response_key != expected_key:
                await _raise_handshake_error("Invalid challenge response")

            conn = resp.connection
            assert conn is not None
            conn_proto = conn.protocol
            assert conn_proto is not None
            conn_proto.read_timeout = None

            transport = conn.transport
            assert transport is not None
            web_socket_data_queue = cast(Callable[..., Any], getattr(aiohttp.client_ws, "WebSocketDataQueue"))
            reader = web_socket_data_queue(conn_proto, 2**16, loop=session._loop)
            conn_proto.set_parser(WebSocketReader(reader, max_msg_size), reader)
            writer = WebSocketWriter(conn_proto, transport, use_mask=True, compress=0, notakeover=False)
        except BaseException as exc:
            if circuit_breaker is not None and not _cb_recorded and isinstance(exc, Exception):
                await circuit_breaker._record_failure(exc)
                _cb_recorded = True
            resp.close()
            raise

        websocket = session._ws_response_class(
            reader,
            writer,
            None,
            resp,
            DEFAULT_WS_CLIENT_TIMEOUT,
            True,
            True,
            session._loop,
            heartbeat=None,
            compress=0,
            client_notakeover=False,
        )
        if hold_half_open_probe and is_probe and circuit_breaker is not None:
            _bind_half_open_probe(websocket, circuit_breaker)
        return websocket, websocket
    finally:
        if is_probe and circuit_breaker is not None and not hold_half_open_probe:
            await circuit_breaker.release_half_open_probe()


async def _stream_websocket_events(
    websocket: aiohttp.ClientWebSocketResponse,
    *,
    idle_timeout_seconds: float,
    total_timeout_seconds: float | None,
    max_event_bytes: int,
) -> AsyncIterator[str]:
    deadline = None if total_timeout_seconds is None else time.monotonic() + total_timeout_seconds

    while True:
        timeout_seconds = idle_timeout_seconds
        if deadline is not None:
            remaining = deadline - time.monotonic()
            if remaining <= 0:
                raise asyncio.TimeoutError
            timeout_seconds = min(timeout_seconds, remaining)

        try:
            message = await asyncio.wait_for(websocket.receive(), timeout=timeout_seconds)
        except asyncio.TimeoutError as exc:
            if deadline is not None and deadline - time.monotonic() <= 0:
                raise
            raise StreamIdleTimeoutError() from exc

        if message.type in {aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSED}:
            break
        if message.type == aiohttp.WSMsgType.ERROR:
            exc = websocket.exception()
            if exc is None and isinstance(message.data, BaseException):
                exc = message.data
            raise exc or aiohttp.ClientError("Upstream websocket error")
        if message.type not in {aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.BINARY}:
            continue

        if message.type == aiohttp.WSMsgType.TEXT:
            text = message.data
        else:
            text = message.data.decode("utf-8", errors="replace")
        text_bytes = text.encode("utf-8")
        if len(text_bytes) > max_event_bytes:
            raise StreamEventTooLargeError(len(text_bytes), max_event_bytes)

        try:
            payload = json.loads(text)
        except json.JSONDecodeError:
            continue
        if not isinstance(payload, dict):
            continue
        normalized = _normalize_stream_event_payload(payload)
        event_type = payload.get("type")
        yield format_sse_event(normalized)
        if isinstance(event_type, str) and event_type in (
            "response.completed",
            "response.failed",
            "response.incomplete",
        ):
            break


async def _stream_responses_via_websocket(
    *,
    payload_dict: JsonObject,
    url: str,
    headers: Mapping[str, str],
    client_session: aiohttp.ClientSession,
    effective_total_timeout: float,
    effective_connect_timeout: float,
    effective_idle_timeout: float,
    max_event_bytes: int,
    raise_for_status: bool,
    account_id: str | None = None,
) -> AsyncIterator[str]:
    websocket_url = _to_websocket_upstream_url(url)
    request_started_at = time.monotonic()
    request_payload = _build_websocket_response_create_payload(payload_dict)
    websocket_cm: AsyncContextManager[aiohttp.ClientWebSocketResponse] | None = None
    websocket: aiohttp.ClientWebSocketResponse | None = None
    circuit_breaker = None
    lifecycle_recorded = False
    seen_terminal = False
    settings = get_settings()
    if account_id is not None:
        circuit_breaker = get_circuit_breaker_for_account(account_id, settings)

    async def _record_lifecycle_success() -> None:
        nonlocal lifecycle_recorded
        if circuit_breaker is None or lifecycle_recorded:
            return
        await circuit_breaker._record_success()
        lifecycle_recorded = True

    async def _record_lifecycle_failure(exc: Exception) -> None:
        nonlocal lifecycle_recorded
        if circuit_breaker is None or lifecycle_recorded:
            return
        await circuit_breaker._record_failure(exc)
        lifecycle_recorded = True

    connect_timeout_seconds = min(
        effective_connect_timeout,
        _remaining_total_timeout(effective_total_timeout, request_started_at, time.monotonic())
        or effective_connect_timeout,
    )
    websocket_cm, websocket = await _open_upstream_websocket(
        session=client_session,
        url=websocket_url,
        headers=headers,
        connect_timeout_seconds=connect_timeout_seconds,
        max_msg_size=max_event_bytes,
        account_id=account_id,
        hold_half_open_probe=True,
    )

    try:
        send_json = getattr(websocket, "send_json", None)
        remaining_total_timeout = _remaining_total_timeout(
            effective_total_timeout,
            request_started_at,
            time.monotonic(),
        )
        if callable(send_json):
            await asyncio.wait_for(
                cast(Any, send_json)(request_payload),
                timeout=remaining_total_timeout,
            )
        else:
            await asyncio.wait_for(
                websocket.send_str(json.dumps(request_payload, ensure_ascii=True, separators=(",", ":"))),
                timeout=remaining_total_timeout,
            )
        remaining_total_timeout = _remaining_total_timeout(
            effective_total_timeout,
            request_started_at,
            time.monotonic(),
        )
        async for event in _stream_websocket_events(
            websocket,
            idle_timeout_seconds=effective_idle_timeout,
            total_timeout_seconds=remaining_total_timeout,
            max_event_bytes=max_event_bytes,
        ):
            parsed_event = parse_sse_event(event)
            if parsed_event and parsed_event.type in ("response.completed", "response.failed", "response.incomplete"):
                seen_terminal = True
                await _record_lifecycle_success()
            yield event
        if not seen_terminal:
            await _record_lifecycle_failure(aiohttp.ClientError("Upstream websocket closed without terminal event"))
    except Exception as exc:
        await _record_lifecycle_failure(exc)
        raise
    finally:
        try:
            if websocket_cm is not None:
                await websocket_cm.__aexit__(None, None, None)
        finally:
            await _release_bound_half_open_probe(websocket)


def _build_websocket_response_create_payload(payload_dict: JsonObject) -> JsonObject:
    request_payload: JsonObject = {
        key: value for key, value in payload_dict.items() if key not in _WEBSOCKET_RESPONSE_CREATE_EXCLUDED_FIELDS
    }
    request_payload["type"] = "response.create"
    return request_payload


async def _inline_input_image_urls(
    payload: JsonObject,
    session: "ImageFetchSession",
    connect_timeout: float,
) -> dict[str, JsonValue]:
    payload_dict = dict(payload)
    input_value = payload_dict.get("input")
    if not isinstance(input_value, list):
        return payload_dict
    updated_input: list[JsonValue] = []
    changed = False
    for item in input_value:
        if not isinstance(item, dict):
            updated_input.append(item)
            continue
        content = item.get("content")
        updated_content, content_changed = await _inline_content_images(content, session, connect_timeout)
        if content_changed:
            new_item = dict(item)
            new_item["content"] = updated_content
            updated_input.append(new_item)
            changed = True
        else:
            updated_input.append(item)
    if not changed:
        return payload_dict
    payload_dict["input"] = updated_input
    return payload_dict


async def _inline_content_images(
    content: JsonValue,
    session: "ImageFetchSession",
    connect_timeout: float,
) -> tuple[JsonValue, bool]:
    if content is None:
        return content, False
    parts = content if isinstance(content, list) else [content]
    updated_parts: list[JsonValue] = []
    changed = False
    for part in parts:
        if not isinstance(part, dict):
            updated_parts.append(part)
            continue
        part_type = part.get("type")
        image_url = part.get("image_url") if part_type == "input_image" else None
        if isinstance(image_url, str) and image_url.startswith(("http://", "https://")):
            data_url = await _fetch_image_data_url(session, image_url, connect_timeout)
            if data_url:
                new_part = dict(part)
                new_part["image_url"] = data_url
                updated_parts.append(new_part)
                changed = True
                continue
        updated_parts.append(part)
    if isinstance(content, list):
        return updated_parts, changed
    return (updated_parts[0] if updated_parts else ""), changed


async def _fetch_image_data_url(
    session: "ImageFetchSession",
    image_url: str,
    connect_timeout: float,
) -> str | None:
    target = await _resolve_safe_image_fetch_target(image_url, connect_timeout=connect_timeout)
    if target is None:
        return None
    timeout = aiohttp.ClientTimeout(
        total=_IMAGE_INLINE_TIMEOUT_SECONDS,
        sock_connect=connect_timeout,
        sock_read=_IMAGE_INLINE_TIMEOUT_SECONDS,
    )
    headers = {"Host": target.host_header}
    for request_url in target.request_urls:
        try:
            async with session.get(
                request_url,
                timeout=timeout,
                allow_redirects=False,
                headers=headers,
                server_hostname=target.server_hostname,
            ) as resp:
                if resp.status != 200:
                    continue
                content_type = resp.headers.get("Content-Type")
                mime_type = content_type.split(";", 1)[0].strip() if isinstance(content_type, str) else ""
                if not mime_type:
                    mime_type = "application/octet-stream"
                data = bytearray()
                async for chunk in resp.content.iter_chunked(_IMAGE_INLINE_CHUNK_SIZE):
                    if not chunk:
                        continue
                    data.extend(chunk)
                    if len(data) > _IMAGE_INLINE_MAX_BYTES:
                        return None
                if not data:
                    continue
                encoded = base64.b64encode(data).decode("ascii")
                return f"data:{mime_type};base64,{encoded}"
        except (aiohttp.ClientError, asyncio.TimeoutError):
            continue
    return None


@dataclass(slots=True, frozen=True)
class SafeImageFetchTarget:
    request_urls: tuple[str, ...]
    host_header: str
    server_hostname: str


def _build_pinned_request_url(parsed: ParseResult, resolved_ip: str) -> str:
    path = parsed.path or "/"
    ip_host = f"[{resolved_ip}]" if ":" in resolved_ip else resolved_ip
    try:
        parsed_port = parsed.port
    except ValueError:
        parsed_port = None
    netloc = f"{ip_host}:{parsed_port}" if parsed_port is not None else ip_host
    return urlunparse((parsed.scheme, netloc, path, parsed.params, parsed.query, parsed.fragment))


async def _resolve_safe_image_fetch_target(
    url: str,
    *,
    connect_timeout: float,
) -> SafeImageFetchTarget | None:
    settings = get_settings()
    if not settings.image_inline_fetch_enabled:
        return None

    parsed = urlparse(url)
    if parsed.scheme != "https":
        return None
    if parsed.username or parsed.password:
        return None
    hostname = parsed.hostname
    if not hostname:
        return None
    host = hostname.strip().lower().rstrip(".")
    if not host:
        return None
    if host in _BLOCKED_LITERAL_HOSTS:
        return None

    allowed_hosts = settings.image_inline_allowed_hosts
    if allowed_hosts and host not in allowed_hosts:
        return None

    literal_ip = _parse_ip_literal(host)
    if literal_ip is not None:
        if _is_disallowed_ip(literal_ip):
            return None
        resolved_ips = [literal_ip.compressed]
    else:
        resolve_timeout = min(connect_timeout, _IMAGE_INLINE_TIMEOUT_SECONDS)
        resolved_ips = await _resolve_global_ips(host, timeout_seconds=resolve_timeout)
        if not resolved_ips:
            return None

    request_urls = tuple(_build_pinned_request_url(parsed, resolved_ip) for resolved_ip in resolved_ips)
    if not request_urls:
        return None

    try:
        parsed_port = parsed.port
    except ValueError:
        return None
    host_header = host if parsed_port in (None, 443) else f"{host}:{parsed_port}"
    return SafeImageFetchTarget(
        request_urls=request_urls,
        host_header=host_header,
        server_hostname=host,
    )


async def _is_safe_image_fetch_url(url: str, *, connect_timeout: float) -> bool:
    target = await _resolve_safe_image_fetch_target(url, connect_timeout=connect_timeout)
    return target is not None


def _parse_ip_literal(host: str) -> ipaddress.IPv4Address | ipaddress.IPv6Address | None:
    try:
        return ipaddress.ip_address(host)
    except ValueError:
        return None


def _is_blocked_ip_literal(host: str) -> bool:
    ip = _parse_ip_literal(host)
    if ip is None:
        return False
    return _is_disallowed_ip(ip)


async def _resolve_global_ips(host: str, *, timeout_seconds: float) -> list[str] | None:
    loop = asyncio.get_running_loop()
    try:
        infos = await asyncio.wait_for(
            loop.getaddrinfo(host, None, proto=socket.IPPROTO_TCP),
            timeout=timeout_seconds,
        )
    except (OSError, asyncio.TimeoutError):
        return None
    if not infos:
        return None

    resolved_ips: list[str] = []
    seen: set[str] = set()
    for info in infos:
        sockaddr = info[4]
        if not sockaddr:
            return None
        addr = sockaddr[0]
        ip = _parse_ip_literal(addr)
        if ip is None:
            return None
        if _is_disallowed_ip(ip):
            return None
        normalized_ip = ip.compressed
        if normalized_ip in seen:
            continue
        seen.add(normalized_ip)
        resolved_ips.append(normalized_ip)
    return resolved_ips or None


async def _resolves_to_blocked_ip(host: str, *, timeout_seconds: float) -> bool:
    resolved_ips = await _resolve_global_ips(host, timeout_seconds=timeout_seconds)
    return resolved_ips is None


def _is_disallowed_ip(ip: ipaddress.IPv4Address | ipaddress.IPv6Address) -> bool:
    if ip.is_multicast:
        return True
    return not ip.is_global


class ImageFetchContent(Protocol):
    def iter_chunked(self, size: int) -> AsyncIterator[bytes]: ...


class ImageFetchResponse(Protocol):
    status: int
    headers: Mapping[str, str]
    content: ImageFetchContent


class ImageFetchSession(Protocol):
    def get(
        self,
        url: str,
        timeout: aiohttp.ClientTimeout,
        *,
        allow_redirects: bool = False,
        headers: Mapping[str, str] | None = None,
        server_hostname: str | None = None,
    ) -> AsyncContextManager[ImageFetchResponse]: ...


def _as_image_fetch_session(session: aiohttp.ClientSession) -> ImageFetchSession:
    return cast(ImageFetchSession, session)


async def stream_responses(
    payload: ResponsesRequest,
    headers: Mapping[str, str],
    access_token: str,
    account_id: str | None,
    base_url: str | None = None,
    raise_for_status: bool = False,
    session: aiohttp.ClientSession | None = None,
    upstream_stream_transport_override: str | None = None,
) -> AsyncIterator[str]:
    settings = get_settings()
    upstream_base = (base_url or settings.upstream_base_url).rstrip("/")
    url = f"{upstream_base}/codex/responses"
    pre_request_started_at = time.monotonic()
    # Keep a default total timeout so direct callers cannot hang forever before
    # response headers or the first SSE event. ProxyService stream attempts clamp
    # this further by installing per-attempt overrides from the remaining budget.
    request_total_timeout = _effective_stream_timeout(
        settings.proxy_request_budget_seconds,
        "total",
    )
    effective_connect_timeout = _effective_stream_timeout(settings.upstream_connect_timeout_seconds, "connect")
    effective_idle_timeout = _effective_stream_timeout(settings.stream_idle_timeout_seconds, "idle")

    seen_terminal = False
    status_code: int | None = None
    error_code: str | None = None
    error_message: str | None = None
    client_session = session or get_http_client().session
    payload_dict = _normalize_chatgpt_web_payload(payload.to_payload())
    if settings.image_inline_fetch_enabled:
        payload_dict = await _inline_input_image_urls(
            payload_dict,
            _as_image_fetch_session(client_session),
            effective_connect_timeout,
        )
    transport_mode = _configured_stream_transport(
        transport=settings.upstream_stream_transport,
        transport_override=upstream_stream_transport_override,
    )
    transport = _resolve_stream_transport(
        transport=settings.upstream_stream_transport,
        transport_override=upstream_stream_transport_override,
        model=payload.model,
        headers=headers,
    )
    if transport == "websocket":
        upstream_headers = _build_upstream_websocket_headers(headers, access_token, account_id)
        method = "GET"
    else:
        upstream_headers = _build_upstream_headers(headers, access_token, account_id)
        method = "POST"
    remaining_request_timeout = _remaining_total_timeout(
        request_total_timeout,
        pre_request_started_at,
        time.monotonic(),
    )
    timeout = aiohttp.ClientTimeout(
        total=remaining_request_timeout,
        sock_connect=effective_connect_timeout,
        sock_read=None,
    )
    started_at = time.monotonic()

    async def _stream_via_http(
        current_headers: Mapping[str, str],
        current_timeout: aiohttp.ClientTimeout,
    ) -> AsyncIterator[str]:
        nonlocal status_code, error_code, error_message, seen_terminal

        async with _service_circuit_breaker_context(
            client_session.post(
                url,
                json=payload_dict,
                headers=current_headers,
                timeout=current_timeout,
            ),
            settings=settings,
            account_id=account_id,
        ) as resp:
            status_code = resp.status
            if resp.status >= 400:
                if raise_for_status:
                    error_payload = await _error_payload_from_response(resp)
                    error_code, error_message = _error_details_from_envelope(error_payload)
                    raise ProxyResponseError(resp.status, error_payload)
                event = await _error_event_from_response(resp)
                error_code, error_message = _error_details_from_failed_event(event)
                yield format_sse_event(event)
                return

            async for event_block in _iter_sse_events(
                resp,
                effective_idle_timeout,
                settings.max_sse_event_bytes,
            ):
                event_block = _normalize_sse_event_block(event_block)
                event = parse_sse_event(event_block)
                if event:
                    event_type = event.type
                    if event_type in ("response.completed", "response.failed", "response.incomplete"):
                        seen_terminal = True
                yield event_block
                if seen_terminal:
                    break

    _maybe_log_upstream_request_start(
        kind="responses",
        url=url,
        headers=upstream_headers,
        method=method,
        payload_summary=_summarize_json_payload(payload_dict),
        payload_json=json.dumps(payload_dict, ensure_ascii=True, separators=(",", ":"))
        if settings.log_upstream_request_payload
        else None,
    )
    try:
        if transport == "websocket":
            try:
                async for event_block in _stream_responses_via_websocket(
                    payload_dict=payload_dict,
                    url=url,
                    headers=upstream_headers,
                    client_session=client_session,
                    effective_total_timeout=(remaining_request_timeout or settings.proxy_request_budget_seconds),
                    effective_connect_timeout=effective_connect_timeout,
                    effective_idle_timeout=effective_idle_timeout,
                    max_event_bytes=settings.max_sse_event_bytes,
                    raise_for_status=raise_for_status,
                    account_id=account_id,
                ):
                    if status_code is None:
                        status_code = 101
                    event = parse_sse_event(event_block)
                    if event:
                        event_type = event.type
                        if event_type in ("response.completed", "response.failed", "response.incomplete"):
                            seen_terminal = True
                    yield event_block
                    if seen_terminal:
                        break
            except aiohttp.WSServerHandshakeError as exc:
                if not _should_fallback_to_http_after_websocket_handshake_error(transport_mode, exc):
                    error_payload = _error_payload_from_websocket_handshake_error(exc)
                    error_code, error_message = _error_details_from_envelope(error_payload)
                    status_code = exc.status
                    if error_message is None:
                        error_message = exc.message or str(exc)
                    if error_code is None:
                        error_code = "upstream_error"
                    if raise_for_status:
                        raise ProxyResponseError(exc.status, error_payload) from exc
                    yield format_sse_event(
                        response_failed_event(error_code, error_message, response_id=get_request_id())
                    )
                    return

                logger.warning(
                    "upstream_websocket_handshake_rejected request_id=%s status=%s target=%s retrying_transport=http",
                    get_request_id(),
                    exc.status,
                    _summarize_upstream_target(url),
                )
                _maybe_log_upstream_request_complete(
                    kind="responses",
                    url=url,
                    headers=upstream_headers,
                    method=method,
                    started_at=started_at,
                    status_code=exc.status,
                    error_code="upstream_websocket_handshake_rejected",
                    error_message=str(exc),
                )

                transport = "http"
                upstream_headers = _build_upstream_headers(headers, access_token, account_id)
                method = "POST"
                remaining_request_timeout = _remaining_total_timeout(
                    request_total_timeout,
                    pre_request_started_at,
                    time.monotonic(),
                )
                timeout = aiohttp.ClientTimeout(
                    total=remaining_request_timeout,
                    sock_connect=effective_connect_timeout,
                    sock_read=None,
                )
                started_at = time.monotonic()
                _maybe_log_upstream_request_start(
                    kind="responses",
                    url=url,
                    headers=upstream_headers,
                    method=method,
                    payload_summary=_summarize_json_payload(payload_dict),
                    payload_json=json.dumps(payload_dict, ensure_ascii=True, separators=(",", ":"))
                    if settings.log_upstream_request_payload
                    else None,
                )
                async for event_block in _stream_via_http(upstream_headers, timeout):
                    yield event_block
        else:
            async for event_block in _stream_via_http(upstream_headers, timeout):
                yield event_block
    except ProxyResponseError as exc:
        status_code = exc.status_code
        raise
    except StreamIdleTimeoutError:
        error_code = "stream_idle_timeout"
        error_message = "Upstream stream idle timeout"
        yield format_sse_event(
            response_failed_event(
                "stream_idle_timeout",
                "Upstream stream idle timeout",
                response_id=get_request_id(),
            ),
        )
        return
    except StreamEventTooLargeError as exc:
        error_code = "stream_event_too_large"
        error_message = str(exc)
        yield format_sse_event(
            response_failed_event(
                "stream_event_too_large",
                str(exc),
                response_id=get_request_id(),
            ),
        )
        return
    except CircuitBreakerOpenError:
        error_code = "upstream_unavailable"
        error_message = "Upstream circuit breaker is open"
        yield format_sse_event(
            response_failed_event("upstream_unavailable", error_message, response_id=get_request_id()),
        )
        return
    except aiohttp.ClientError as exc:
        error_code = "upstream_unavailable"
        error_message = str(exc)
        yield format_sse_event(
            response_failed_event("upstream_unavailable", str(exc), response_id=get_request_id()),
        )
        return
    except asyncio.CancelledError:
        raise
    except asyncio.TimeoutError as exc:
        if isinstance(exc, aiohttp.ClientError):
            error_code = "upstream_unavailable"
            error_message = str(exc) or "Request to upstream timed out"
            yield format_sse_event(
                response_failed_event("upstream_unavailable", error_message, response_id=get_request_id()),
            )
            return
        error_code = "upstream_request_timeout"
        error_message = "Proxy request budget exhausted"
        yield format_sse_event(
            response_failed_event(
                "upstream_request_timeout",
                "Proxy request budget exhausted",
                response_id=get_request_id(),
            ),
        )
        return
    except Exception as exc:
        error_code = "upstream_error"
        error_message = str(exc)
        yield format_sse_event(response_failed_event("upstream_error", str(exc), response_id=get_request_id()))
        return
    else:
        if not seen_terminal:
            error_code = "stream_incomplete"
            error_message = "Upstream closed stream without completion"
            yield format_sse_event(
                response_failed_event(
                    "stream_incomplete",
                    "Upstream closed stream without completion",
                    response_id=get_request_id(),
                ),
            )
    finally:
        _maybe_log_upstream_request_complete(
            kind="responses",
            url=url,
            headers=upstream_headers,
            method=method,
            started_at=started_at,
            status_code=status_code,
            error_code=error_code,
            error_message=error_message,
        )


def push_stream_timeout_overrides(
    *,
    connect_timeout_seconds: float | None = None,
    idle_timeout_seconds: float | None = None,
    total_timeout_seconds: float | None = None,
) -> tuple[
    float | None,
    float | None,
    float | None,
]:
    previous = (
        _STREAM_CONNECT_TIMEOUT_OVERRIDE.get(),
        _STREAM_IDLE_TIMEOUT_OVERRIDE.get(),
        _STREAM_TOTAL_TIMEOUT_OVERRIDE.get(),
    )
    _STREAM_CONNECT_TIMEOUT_OVERRIDE.set(connect_timeout_seconds)
    _STREAM_IDLE_TIMEOUT_OVERRIDE.set(idle_timeout_seconds)
    _STREAM_TOTAL_TIMEOUT_OVERRIDE.set(total_timeout_seconds)
    return previous


def pop_stream_timeout_overrides(
    tokens: tuple[
        float | None,
        float | None,
        float | None,
    ],
) -> None:
    connect_timeout, idle_timeout, total_timeout = tokens
    _STREAM_CONNECT_TIMEOUT_OVERRIDE.set(connect_timeout)
    _STREAM_IDLE_TIMEOUT_OVERRIDE.set(idle_timeout)
    _STREAM_TOTAL_TIMEOUT_OVERRIDE.set(total_timeout)


@contextlib.contextmanager
def override_stream_timeouts(
    *,
    connect_timeout_seconds: float | None = None,
    idle_timeout_seconds: float | None = None,
    total_timeout_seconds: float | None = None,
):
    tokens = push_stream_timeout_overrides(
        connect_timeout_seconds=connect_timeout_seconds,
        idle_timeout_seconds=idle_timeout_seconds,
        total_timeout_seconds=total_timeout_seconds,
    )
    try:
        yield
    finally:
        pop_stream_timeout_overrides(tokens)


def push_compact_timeout_overrides(
    *,
    connect_timeout_seconds: float | None = None,
    total_timeout_seconds: float | None = None,
) -> tuple[contextvars.Token[float | None], contextvars.Token[float | None]]:
    return (
        _COMPACT_CONNECT_TIMEOUT_OVERRIDE.set(connect_timeout_seconds),
        _COMPACT_TOTAL_TIMEOUT_OVERRIDE.set(total_timeout_seconds),
    )


def pop_compact_timeout_overrides(
    tokens: tuple[contextvars.Token[float | None], contextvars.Token[float | None]],
) -> None:
    connect_token, total_token = tokens
    _COMPACT_CONNECT_TIMEOUT_OVERRIDE.reset(connect_token)
    _COMPACT_TOTAL_TIMEOUT_OVERRIDE.reset(total_token)


def push_transcribe_timeout_overrides(
    *,
    connect_timeout_seconds: float | None = None,
    total_timeout_seconds: float | None = None,
) -> tuple[contextvars.Token[float | None], contextvars.Token[float | None]]:
    return (
        _TRANSCRIBE_CONNECT_TIMEOUT_OVERRIDE.set(connect_timeout_seconds),
        _TRANSCRIBE_TOTAL_TIMEOUT_OVERRIDE.set(total_timeout_seconds),
    )


def pop_transcribe_timeout_overrides(
    tokens: tuple[contextvars.Token[float | None], contextvars.Token[float | None]],
) -> None:
    connect_token, total_token = tokens
    _TRANSCRIBE_CONNECT_TIMEOUT_OVERRIDE.reset(connect_token)
    _TRANSCRIBE_TOTAL_TIMEOUT_OVERRIDE.reset(total_token)


async def compact_responses(
    payload: ResponsesCompactRequest,
    headers: Mapping[str, str],
    access_token: str,
    account_id: str | None,
    session: aiohttp.ClientSession | None = None,
) -> CompactResponsePayload:
    transport = _CompactCommandTransport(
        payload=payload,
        headers=headers,
        access_token=access_token,
        account_id=account_id,
        session=session or get_http_client().session,
    )
    return await transport.execute()


def _is_retryable_compact_status(status_code: int) -> bool:
    return status_code in {401, 500, 502, 503, 504}


@dataclass(slots=True)
class _CompactCommandTransport:
    payload: ResponsesCompactRequest
    headers: Mapping[str, str]
    access_token: str
    account_id: str | None
    session: aiohttp.ClientSession

    async def execute(self) -> CompactResponsePayload:
        settings = get_settings()
        upstream_base = settings.upstream_base_url.rstrip("/")
        url = f"{upstream_base}/codex/responses/compact"
        upstream_headers = _build_upstream_headers(
            self.headers,
            self.access_token,
            self.account_id,
            accept="application/json",
        )
        pre_request_started_at = time.monotonic()
        compact_timeout_seconds = _effective_compact_total_timeout(settings.upstream_compact_timeout_seconds)
        effective_connect_timeout = _effective_compact_connect_timeout(settings.upstream_connect_timeout_seconds)
        payload_dict = _normalize_chatgpt_web_payload(self.payload.to_payload())
        if settings.image_inline_fetch_enabled:
            payload_dict = await _inline_input_image_urls(
                payload_dict,
                _as_image_fetch_session(self.session),
                effective_connect_timeout,
            )
        now = time.monotonic()
        compact_timeout_seconds = _remaining_total_timeout(
            compact_timeout_seconds,
            pre_request_started_at,
            now,
        )
        effective_connect_timeout = max(
            0.001,
            _remaining_total_timeout(
                effective_connect_timeout,
                pre_request_started_at,
                now,
            )
            or effective_connect_timeout,
        )
        timeout = aiohttp.ClientTimeout(
            total=compact_timeout_seconds,
            sock_connect=effective_connect_timeout,
            sock_read=compact_timeout_seconds,
        )
        started_at = time.monotonic()
        status_code: int | None = None
        error_code: str | None = None
        error_message: str | None = None
        failure_phase: str | None = None
        payload_object: str | None = None
        failure_detail: str | None = None
        failure_exception_type: str | None = None
        retryable_same_contract: bool | None = None
        _maybe_log_upstream_request_start(
            kind="responses_compact",
            url=url,
            headers=upstream_headers,
            method="POST",
            payload_summary=_summarize_json_payload(payload_dict),
            payload_json=json.dumps(payload_dict, ensure_ascii=True, separators=(",", ":"))
            if settings.log_upstream_request_payload
            else None,
        )
        try:
            async with _service_circuit_breaker_context(
                self.session.post(
                    url,
                    json=payload_dict,
                    headers=upstream_headers,
                    timeout=timeout,
                ),
                settings=settings,
                account_id=self.account_id,
            ) as resp:
                status_code = resp.status
                if resp.status >= 400:
                    error_payload = await _error_payload_from_response(resp)
                    error_code, error_message = _error_details_from_envelope(error_payload)
                    failure_phase = "status"
                    failure_detail = error_message
                    retryable_same_contract = _is_retryable_compact_status(resp.status)
                    raise ProxyResponseError(
                        resp.status,
                        error_payload,
                        failure_phase=failure_phase,
                        retryable_same_contract=retryable_same_contract,
                        failure_detail=failure_detail,
                        upstream_status_code=resp.status,
                    )
                try:
                    data = await resp.json(content_type=None)
                except (aiohttp.ClientError, asyncio.TimeoutError) as exc:
                    message = str(exc) or "Request to upstream timed out"
                    error_code = "upstream_unavailable"
                    error_message = message
                    failure_phase = "body_read"
                    failure_detail = message
                    failure_exception_type = type(exc).__name__
                    retryable_same_contract = True
                    raise ProxyResponseError(
                        502,
                        openai_error("upstream_unavailable", message),
                        failure_phase=failure_phase,
                        retryable_same_contract=retryable_same_contract,
                        failure_detail=failure_detail,
                        failure_exception_type=failure_exception_type,
                        upstream_status_code=resp.status,
                    ) from exc
                except Exception as exc:
                    error_code = "upstream_error"
                    error_message = "Invalid JSON from upstream"
                    failure_phase = "parse"
                    failure_detail = str(exc) or error_message
                    failure_exception_type = type(exc).__name__
                    raise ProxyResponseError(
                        502,
                        openai_error("upstream_error", "Invalid JSON from upstream"),
                        failure_phase=failure_phase,
                        failure_detail=failure_detail,
                        failure_exception_type=failure_exception_type,
                        upstream_status_code=resp.status,
                    ) from exc
                parsed = parse_compact_response_payload(data)
                if parsed:
                    payload_object = parsed.object
                    return parsed
                error_code = "upstream_error"
                error_message = "Unexpected upstream payload"
                failure_phase = "parse"
                failure_detail = f"payload_type={type(data).__name__}"
                raise ProxyResponseError(
                    502,
                    openai_error("upstream_error", "Unexpected upstream payload"),
                    failure_phase=failure_phase,
                    failure_detail=failure_detail,
                    upstream_status_code=resp.status,
                )
        except ProxyResponseError as exc:
            if error_code is None and error_message is None:
                error_code, error_message = _error_details_from_envelope(exc.payload)
            failure_phase = failure_phase or exc.failure_phase
            failure_detail = failure_detail or exc.failure_detail
            failure_exception_type = failure_exception_type or exc.failure_exception_type
            if retryable_same_contract is None:
                retryable_same_contract = exc.retryable_same_contract
            raise
        except CircuitBreakerOpenError as exc:
            error_code = "upstream_unavailable"
            error_message = "Upstream circuit breaker is open"
            failure_phase = "connect"
            failure_detail = str(exc)
            failure_exception_type = type(exc).__name__
            retryable_same_contract = True
            raise ProxyResponseError(
                503,
                openai_error("upstream_unavailable", error_message),
                failure_phase=failure_phase,
                retryable_same_contract=retryable_same_contract,
                failure_detail=failure_detail,
                failure_exception_type=failure_exception_type,
            ) from exc
        except (aiohttp.ClientError, asyncio.TimeoutError) as exc:
            message = str(exc) or "Request to upstream timed out"
            error_code = "upstream_unavailable"
            error_message = message
            failure_phase = "connect"
            failure_detail = message
            failure_exception_type = type(exc).__name__
            retryable_same_contract = True
            raise ProxyResponseError(
                502,
                openai_error("upstream_unavailable", message),
                failure_phase=failure_phase,
                retryable_same_contract=retryable_same_contract,
                failure_detail=failure_detail,
                failure_exception_type=failure_exception_type,
            ) from exc
        finally:
            _maybe_log_upstream_request_complete(
                kind="responses_compact",
                url=url,
                headers=upstream_headers,
                method="POST",
                started_at=started_at,
                status_code=status_code,
                error_code=error_code,
                error_message=error_message,
                failure_phase=failure_phase,
                payload_object=payload_object,
                failure_detail=failure_detail,
                failure_exception_type=failure_exception_type,
                retryable_same_contract=retryable_same_contract,
            )


async def transcribe_audio(
    audio_bytes: bytes,
    *,
    filename: str,
    content_type: str | None,
    prompt: str | None,
    headers: Mapping[str, str],
    access_token: str,
    account_id: str | None,
    base_url: str | None = None,
    session: aiohttp.ClientSession | None = None,
) -> dict[str, JsonValue]:
    settings = get_settings()
    upstream_base = (base_url or settings.upstream_base_url).rstrip("/")
    url = f"{upstream_base}/transcribe"
    upstream_headers = _build_upstream_transcribe_headers(
        headers,
        access_token,
        account_id,
    )

    effective_total_timeout = _effective_transcribe_total_timeout(
        settings.transcription_request_budget_seconds,
    )
    effective_connect_timeout = _effective_transcribe_connect_timeout(settings.upstream_connect_timeout_seconds)
    timeout = aiohttp.ClientTimeout(
        total=effective_total_timeout,
        sock_connect=effective_connect_timeout,
        sock_read=effective_total_timeout,
    )

    normalized_filename = filename.strip() if filename else ""
    if not normalized_filename:
        normalized_filename = "audio.wav"
    normalized_content_type = content_type.strip() if content_type else ""
    if not normalized_content_type:
        normalized_content_type = "application/octet-stream"

    form = aiohttp.FormData()
    form.add_field(
        "file",
        audio_bytes,
        filename=normalized_filename,
        content_type=normalized_content_type,
    )
    if prompt is not None:
        form.add_field("prompt", prompt)

    client_session = session or get_http_client().session
    started_at = time.monotonic()
    status_code: int | None = None
    error_code: str | None = None
    error_message: str | None = None
    metadata = _summarize_transcription_payload(
        filename=normalized_filename,
        content_type=normalized_content_type,
        prompt=prompt,
        audio_bytes=audio_bytes,
    )
    _maybe_log_upstream_request_start(
        kind="transcribe",
        url=url,
        headers=upstream_headers,
        method="POST",
        payload_summary=json.dumps(metadata, ensure_ascii=True, separators=(",", ":")),
        payload_json=json.dumps(metadata, ensure_ascii=True, separators=(",", ":"))
        if settings.log_upstream_request_payload
        else None,
    )
    try:
        async with _service_circuit_breaker_context(
            client_session.post(
                url,
                data=form,
                headers=upstream_headers,
                timeout=timeout,
            ),
            settings=settings,
            account_id=account_id,
        ) as resp:
            status_code = resp.status
            if resp.status >= 400:
                error_payload = await _error_payload_from_response(resp)
                error_code, error_message = _error_details_from_envelope(error_payload)
                raise ProxyResponseError(resp.status, error_payload)
            try:
                data = await resp.json(content_type=None)
            except (aiohttp.ClientError, asyncio.TimeoutError) as exc:
                message = str(exc) or "Request to upstream timed out"
                error_code = "upstream_unavailable"
                error_message = message
                raise ProxyResponseError(
                    502,
                    openai_error("upstream_unavailable", message),
                ) from exc
            except Exception as exc:
                error_code = "upstream_error"
                error_message = "Invalid JSON from upstream"
                raise ProxyResponseError(
                    502,
                    openai_error("upstream_error", "Invalid JSON from upstream"),
                ) from exc
            if isinstance(data, dict):
                return data
            raise ProxyResponseError(
                502,
                openai_error("upstream_error", "Unexpected upstream payload"),
            )
    except ProxyResponseError as exc:
        if error_code is None and error_message is None:
            error_code, error_message = _error_details_from_envelope(exc.payload)
        raise
    except CircuitBreakerOpenError as exc:
        error_code = "upstream_unavailable"
        error_message = "Upstream circuit breaker is open"
        raise ProxyResponseError(
            503,
            openai_error("upstream_unavailable", error_message),
        ) from exc
    except (aiohttp.ClientError, asyncio.TimeoutError) as exc:
        message = str(exc) or "Request to upstream timed out"
        error_code = "upstream_unavailable"
        error_message = message
        raise ProxyResponseError(
            502,
            openai_error("upstream_unavailable", message),
        ) from exc
    finally:
        _maybe_log_upstream_request_complete(
            kind="transcribe",
            url=url,
            headers=upstream_headers,
            method="POST",
            started_at=started_at,
            status_code=status_code,
            error_code=error_code,
            error_message=error_message,
        )



from __future__ import annotations

import asyncio
from collections.abc import Collection
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

import pytest

from app.core.auth.refresh import RefreshError
from app.core.crypto import TokenEncryptor
from app.core.usage.models import UsagePayload
from app.db.models import Account, AccountStatus, UsageHistory
from app.modules.usage.additional_quota_keys import canonicalize_additional_quota_key
from app.modules.usage.updater import UsageUpdater, _last_successful_refresh

pytestmark = pytest.mark.unit


@pytest.fixture(autouse=True)
def _clear_refresh_cache():
    """Clear the module-level freshness cache between tests."""
    _last_successful_refresh.clear()
    yield
    _last_successful_refresh.clear()


@dataclass(frozen=True, slots=True)
class UsageEntry:
    account_id: str
    used_percent: float
    input_tokens: int | None
    output_tokens: int | None
    recorded_at: datetime | None
    window: str | None
    reset_at: int | None
    window_minutes: int | None
    credits_has: bool | None
    credits_unlimited: bool | None
    credits_balance: float | None


class StubUsageRepository:
    def __init__(self, *, return_rows: bool = False) -> None:
        self.entries: list[UsageEntry] = []
        self._return_rows = return_rows
        self._next_id = 1

    async def latest_entry_for_account(
        self,
        account_id: str,
        *,
        window: str | None = None,
    ) -> UsageHistory | None:
        for entry in reversed(self.entries):
            normalized_window = entry.window or "primary"
            expected_window = window or "primary"
            if entry.account_id == account_id and normalized_window == expected_window:
                return UsageHistory(
                    id=self._next_id,
                    account_id=entry.account_id,
                    used_percent=entry.used_percent,
                    input_tokens=entry.input_tokens,
                    output_tokens=entry.output_tokens,
                    recorded_at=entry.recorded_at or datetime.now(tz=timezone.utc),
                    window=entry.window,
                    reset_at=entry.reset_at,
                    window_minutes=entry.window_minutes,
                    credits_has=entry.credits_has,
                    credits_unlimited=entry.credits_unlimited,
                    credits_balance=entry.credits_balance,
                )
        return None

    async def add_entry(
        self,
        account_id: str,
        used_percent: float,
        input_tokens: int | None = None,
        output_tokens: int | None = None,
        recorded_at: datetime | None = None,
        window: str | None = None,
        reset_at: int | None = None,
        window_minutes: int | None = None,
        credits_has: bool | None = None,
        credits_unlimited: bool | None = None,
        credits_balance: float | None = None,
    ) -> UsageHistory | None:
        self.entries.append(
            UsageEntry(
                account_id=account_id,
                used_percent=used_percent,
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                recorded_at=recorded_at,
                window=window,
                reset_at=reset_at,
                window_minutes=window_minutes,
                credits_has=credits_has,
                credits_unlimited=credits_unlimited,
                credits_balance=credits_balance,
            )
        )
        if not self._return_rows:
            return None
        entry = UsageHistory(
            id=self._next_id,
            account_id=account_id,
            used_percent=used_percent,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            recorded_at=recorded_at or datetime.now(tz=timezone.utc),
            window=window,
            reset_at=reset_at,
            window_minutes=window_minutes,
            credits_has=credits_has,
            credits_unlimited=credits_unlimited,
            credits_balance=credits_balance,
        )
        self._next_id += 1
        return entry


@dataclass(frozen=True, slots=True)
class AdditionalUsageEntry:
    account_id: str
    limit_name: str
    metered_feature: str
    window: str
    used_percent: float
    reset_at: int | None
    window_minutes: int | None
    quota_key: str | None = None


class StubAdditionalUsageRepository:
    def __init__(self) -> None:
        self.entries: list[AdditionalUsageEntry] = []
        self.deleted_account_ids: list[str] = []
        self.deleted_account_limit_pairs: list[tuple[str, str]] = []
        self.deleted_account_limit_windows: list[tuple[str, str, str]] = []
        self._written_accounts: set[str] = set()

    async def add_entry(
        self,
        account_id: str,
        limit_name: str,
        metered_feature: str,
        window: str,
        used_percent: float,
        reset_at: int | None = None,
        window_minutes: int | None = None,
        recorded_at: datetime | None = None,
        quota_key: str | None = None,
    ) -> None:
        self._written_accounts.add(account_id)
        self.entries.append(
            AdditionalUsageEntry(
                account_id=account_id,
                quota_key=quota_key
                or canonicalize_additional_quota_key(
                    limit_name=limit_name,
                    metered_feature=metered_feature,
                ),
                limit_name=limit_name,
                metered_feature=metered_feature,
                window=window,
                used_percent=used_percent,
                reset_at=reset_at,
                window_minutes=window_minutes,
            )
        )

    async def delete_for_account(self, account_id: str) -> None:
        self.deleted_account_ids.append(account_id)

    async def delete_for_account_and_limit(self, account_id: str, limit_name: str) -> None:
        self.deleted_account_limit_pairs.append((account_id, limit_name))

    async def delete_for_account_and_quota_key(self, account_id: str, quota_key: str) -> None:
        self.deleted_account_limit_pairs.append((account_id, quota_key))

    async def delete_for_account_limit_window(self, account_id: str, limit_name: str, window: str) -> None:
        self.deleted_account_limit_windows.append((account_id, limit_name, window))

    async def delete_for_account_quota_key_window(self, account_id: str, quota_key: str, window: str) -> None:
        self.deleted_account_limit_windows.append((account_id, quota_key, window))

    async def latest_recorded_at_for_account(self, account_id: str):
        from app.core.utils.time import utcnow

        return utcnow() if account_id in self._written_accounts else None

    async def list_limit_names(
        self,
        *,
        account_ids: Collection[str] | None = None,
        since: datetime | None = None,
    ) -> list[str]:
        if account_ids is None:
            return sorted({entry.limit_name for entry in self.entries})
        return sorted({entry.limit_name for entry in self.entries if entry.account_id in account_ids})

    async def list_quota_keys(
        self,
        *,
        account_ids: Collection[str] | None = None,
        since: datetime | None = None,
    ) -> list[str]:
        if account_ids is None:
            return sorted(
                {
                    entry.quota_key
                    or canonicalize_additional_quota_key(
                        limit_name=entry.limit_name,
                        metered_feature=entry.metered_feature,
                    )
                    or entry.limit_name
                    for entry in self.entries
                }
            )
        return sorted(
            {
                entry.quota_key
                or canonicalize_additional_quota_key(
                    limit_name=entry.limit_name,
                    metered_feature=entry.metered_feature,
                )
                or entry.limit_name
                for entry in self.entries
                if entry.account_id in account_ids
            }
        )


def _make_account(account_id: str, chatgpt_account_id: str, email: str = "a@example.com") -> Account:
    encryptor = TokenEncryptor()
    return Account(
        id=account_id,
        chatgpt_account_id=chatgpt_account_id,
        email=email,
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=datetime.now(tz=timezone.utc),
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )


@pytest.mark.asyncio
async def test_usage_updater_includes_chatgpt_account_id_even_when_shared(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    calls: list[dict[str, Any]] = []

    async def stub_fetch_usage(*, access_token: str, account_id: str | None, **_: Any) -> UsagePayload:
        calls.append({"access_token": access_token, "account_id": account_id})
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                    "secondary_window": {
                        "used_percent": 20.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                }
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None)

    shared = "workspace_shared"
    acc_a = _make_account("acc_a", shared, email="a@example.com")
    acc_b = _make_account("acc_b", shared, email="b@example.com")
    acc_c = _make_account("acc_c", "workspace_unique", email="c@example.com")

    await updater.refresh_accounts([acc_a, acc_b, acc_c], latest_usage={})

    assert [call["account_id"] for call in calls] == [shared, shared, "workspace_unique"]


class StubAccountsRepository:
    def __init__(self) -> None:
        self.status_updates: list[dict[str, Any]] = []
        self.token_updates: list[dict[str, Any]] = []
        self.accounts_by_id: dict[str, Account] = {}

    async def get_by_id(self, account_id: str) -> Account | None:
        return self.accounts_by_id.get(account_id)

    async def update_status(
        self,
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
        reset_at: int | None = None,
    ) -> bool:
        account = self.accounts_by_id.get(account_id)
        if account is not None:
            account.status = status
            account.deactivation_reason = deactivation_reason
            account.reset_at = reset_at
        self.status_updates.append(
            {
                "account_id": account_id,
                "status": status,
                "deactivation_reason": deactivation_reason,
            }
        )
        return True

    async def update_tokens(self, *args: Any, **kwargs: Any) -> bool:
        account_id = args[0] if args else kwargs.get("account_id")
        if not isinstance(account_id, str):
            return True
        account = self.accounts_by_id.get(account_id)
        if account is not None:
            account.access_token_encrypted = kwargs["access_token_encrypted"]
            account.refresh_token_encrypted = kwargs["refresh_token_encrypted"]
            account.id_token_encrypted = kwargs["id_token_encrypted"]
            account.last_refresh = kwargs["last_refresh"]
            plan_type = kwargs.get("plan_type")
            email = kwargs.get("email")
            chatgpt_account_id = kwargs.get("chatgpt_account_id")
            if isinstance(plan_type, str):
                account.plan_type = plan_type
            if isinstance(email, str):
                account.email = email
            if isinstance(chatgpt_account_id, str):
                account.chatgpt_account_id = chatgpt_account_id
        self.token_updates.append({"account_id": account_id, **kwargs})
        return True


@pytest.mark.asyncio
async def test_usage_updater_deactivates_on_account_invalid_4xx(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.clients.usage import UsageFetchError
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage_402(**_: Any) -> UsagePayload:
        raise UsageFetchError(402, "Payment Required")

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage_402)

    usage_repo = StubUsageRepository()
    accounts_repo = StubAccountsRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=accounts_repo)

    acc = _make_account("acc_402", "workspace_402", email="payment@example.com")
    accounts_repo.accounts_by_id[acc.id] = acc

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(accounts_repo.status_updates) == 1
    update = accounts_repo.status_updates[0]
    assert update["account_id"] == "acc_402"
    assert update["status"] == AccountStatus.DEACTIVATED
    assert "402" in update["deactivation_reason"]
    assert "Payment Required" in update["deactivation_reason"]


@pytest.mark.asyncio
async def test_usage_updater_does_not_deactivate_on_403(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.clients.usage import UsageFetchError
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage_403(**_: Any) -> UsagePayload:
        raise UsageFetchError(403, "Forbidden")

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage_403)

    usage_repo = StubUsageRepository()
    accounts_repo = StubAccountsRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=accounts_repo)

    acc = _make_account("acc_403", "workspace_403", email="forbidden@example.com")
    accounts_repo.accounts_by_id[acc.id] = acc

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(accounts_repo.status_updates) == 0


@pytest.mark.asyncio
async def test_usage_updater_does_not_deactivate_on_transient_4xx(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.clients.usage import UsageFetchError
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage_429(**_: Any) -> UsagePayload:
        raise UsageFetchError(429, "Too Many Requests")

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage_429)

    usage_repo = StubUsageRepository()
    accounts_repo = StubAccountsRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=accounts_repo)

    acc = _make_account("acc_429", "workspace_429", email="rate@example.com")
    accounts_repo.accounts_by_id[acc.id] = acc

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(accounts_repo.status_updates) == 0


@pytest.mark.asyncio
async def test_usage_updater_does_not_deactivate_on_401(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.clients.usage import UsageFetchError
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage_401(**_: Any) -> UsagePayload:
        raise UsageFetchError(401, "Unauthorized")

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage_401)

    usage_repo = StubUsageRepository()
    accounts_repo = StubAccountsRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=accounts_repo)

    acc = _make_account("acc_401", "workspace_401", email="auth@example.com")
    accounts_repo.accounts_by_id[acc.id] = acc

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(accounts_repo.status_updates) == 0


@pytest.mark.asyncio
async def test_usage_updater_does_not_deactivate_on_5xx(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.clients.usage import UsageFetchError
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage_500(**_: Any) -> UsagePayload:
        raise UsageFetchError(500, "Internal Server Error")

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage_500)

    usage_repo = StubUsageRepository()
    accounts_repo = StubAccountsRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=accounts_repo)

    acc = _make_account("acc_500", "workspace_500", email="server@example.com")
    accounts_repo.accounts_by_id[acc.id] = acc

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(accounts_repo.status_updates) == 0


@pytest.mark.asyncio
async def test_usage_updater_persists_primary_and_secondary_usage(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(*, access_token: str, account_id: str | None, **_: Any) -> UsagePayload:
        assert access_token
        assert account_id == "workspace_123"
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 12.5,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 300,
                    },
                    "secondary_window": {
                        "used_percent": 55.0,
                        "reset_at": 1735693200,
                        "limit_window_seconds": 60,
                    },
                },
                "credits": {"has_credits": True, "unlimited": False, "balance": "42.5"},
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None)
    acc = _make_account("acc_test", "workspace_123", email="persist@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(usage_repo.entries) == 2
    by_window = {entry.window: entry for entry in usage_repo.entries}

    primary = by_window["primary"]
    assert primary.account_id == "acc_test"
    assert primary.used_percent == 12.5
    assert primary.reset_at == 1735689600
    assert primary.window_minutes == 5
    assert primary.credits_has is True
    assert primary.credits_unlimited is False
    assert primary.credits_balance == 42.5

    secondary = by_window["secondary"]
    assert secondary.account_id == "acc_test"
    assert secondary.used_percent == 55.0
    assert secondary.reset_at == 1735693200
    assert secondary.window_minutes == 1
    assert secondary.credits_has is None
    assert secondary.credits_unlimited is None
    assert secondary.credits_balance is None


@pytest.mark.asyncio
async def test_usage_updater_syncs_plan_type_from_usage_payload(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate({"plan_type": "plus"})

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    accounts_repo = StubAccountsRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=accounts_repo)
    acc = _make_account("acc_plan_sync", "workspace_plan_sync", email="plan@example.com")
    accounts_repo.accounts_by_id[acc.id] = acc
    acc.plan_type = "free"

    await updater.refresh_accounts([acc], latest_usage={})

    assert acc.plan_type == "plus"
    assert len(accounts_repo.token_updates) == 1
    token_update = accounts_repo.token_updates[0]
    assert token_update["account_id"] == "acc_plan_sync"
    assert token_update["plan_type"] == "plus"
    assert usage_repo.entries == []


@pytest.mark.asyncio
async def test_usage_updater_computes_reset_at_from_reset_after_seconds(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    monkeypatch.setattr("app.modules.usage.updater._now_epoch", lambda: 1000)

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 1.0,
                        "reset_after_seconds": 120,
                        "limit_window_seconds": 60,
                    }
                }
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None)
    acc = _make_account("acc_reset", "workspace_reset", email="reset@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(usage_repo.entries) == 1
    entry = usage_repo.entries[0]
    assert entry.window == "primary"
    assert entry.reset_at == 1120


@pytest.mark.asyncio
async def test_usage_updater_refresh_accounts_returns_false_when_rate_limit_missing(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate({})

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository(return_rows=True)
    updater = UsageUpdater(usage_repo, accounts_repo=None)
    acc = _make_account("acc_no_rate", "workspace_no_rate", email="no-rate@example.com")

    refreshed = await updater.refresh_accounts([acc], latest_usage={})

    assert refreshed is False
    assert len(usage_repo.entries) == 0


@pytest.mark.asyncio
async def test_usage_updater_refresh_accounts_returns_false_on_401_retry_failure(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.clients.usage import UsageFetchError
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage_401(**_: Any) -> UsagePayload:
        raise UsageFetchError(401, "Unauthorized")

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage_401)

    usage_repo = StubUsageRepository(return_rows=True)
    accounts_repo = StubAccountsRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=accounts_repo)
    assert updater._auth_manager is not None

    async def stub_ensure_fresh(account: Account, *, force: bool = False) -> Account:
        raise RefreshError(code="invalid_grant", message="refresh failed", is_permanent=False)

    monkeypatch.setattr(updater._auth_manager, "ensure_fresh", stub_ensure_fresh)

    acc = _make_account("acc_401_retry", "workspace_401_retry", email="auth-retry@example.com")
    accounts_repo.accounts_by_id[acc.id] = acc
    refreshed = await updater.refresh_accounts([acc], latest_usage={})

    assert refreshed is False
    assert len(usage_repo.entries) == 0


@pytest.mark.parametrize(
    ("primary_used", "secondary_used"),
    [
        (10.0, None),
        (None, 20.0),
    ],
)
@pytest.mark.asyncio
async def test_usage_updater_refresh_accounts_returns_true_when_any_window_written(
    monkeypatch,
    primary_used: float | None,
    secondary_used: float | None,
) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(*, access_token: str, account_id: str | None, **_: Any) -> UsagePayload:
        assert access_token
        assert account_id == "workspace_written"
        rate_limit: dict[str, Any] = {}
        if primary_used is not None:
            rate_limit["primary_window"] = {
                "used_percent": primary_used,
                "reset_at": 1735689600,
                "limit_window_seconds": 60,
            }
        if secondary_used is not None:
            rate_limit["secondary_window"] = {
                "used_percent": secondary_used,
                "reset_at": 1735689600,
                "limit_window_seconds": 60,
            }
        return UsagePayload.model_validate({"rate_limit": rate_limit})

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository(return_rows=True)
    updater = UsageUpdater(usage_repo, accounts_repo=None)
    acc = _make_account("acc_written", "workspace_written", email="written@example.com")

    refreshed = await updater.refresh_accounts([acc], latest_usage={})

    assert refreshed is True
    assert len(usage_repo.entries) == 1


@pytest.mark.asyncio
async def test_usage_updater_refresh_accounts_returns_true_when_partial_write(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(*, account_id: str | None, **_: Any) -> UsagePayload:
        if account_id == "workspace_skip":
            return UsagePayload.model_validate({})
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    }
                }
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository(return_rows=True)
    updater = UsageUpdater(usage_repo, accounts_repo=None)
    acc_skip = _make_account("acc_skip", "workspace_skip", email="skip@example.com")
    acc_write = _make_account("acc_write", "workspace_write", email="write@example.com")

    refreshed = await updater.refresh_accounts([acc_skip, acc_write], latest_usage={})

    assert refreshed is True
    assert len(usage_repo.entries) == 1


@pytest.mark.asyncio
async def test_usage_updater_singleflights_concurrent_refreshes(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    fetch_calls = 0
    fetch_started = asyncio.Event()
    release_fetch = asyncio.Event()

    async def stub_fetch_usage(*, account_id: str | None, **_: Any) -> UsagePayload:
        nonlocal fetch_calls
        fetch_calls += 1
        assert account_id == "workspace_shared_refresh"
        fetch_started.set()
        await release_fetch.wait()
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    }
                }
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository(return_rows=True)
    updater_a = UsageUpdater(usage_repo, accounts_repo=None)
    updater_b = UsageUpdater(usage_repo, accounts_repo=None)
    acc_a = _make_account("acc_singleflight", "workspace_shared_refresh", email="a@example.com")
    acc_b = _make_account("acc_singleflight", "workspace_shared_refresh", email="b@example.com")

    first = asyncio.create_task(updater_a.refresh_accounts([acc_a], latest_usage={}))
    await fetch_started.wait()
    second = asyncio.create_task(updater_b.refresh_accounts([acc_b], latest_usage={}))
    await asyncio.sleep(0.01)

    assert not second.done()

    release_fetch.set()
    first_refreshed, second_refreshed = await asyncio.gather(first, second)

    assert fetch_calls == 1
    assert first_refreshed is True
    assert second_refreshed is True
    assert len(usage_repo.entries) == 1


# --- Additional rate limits tests ---


@pytest.mark.asyncio
async def test_additional_rate_limits_written_to_additional_repo(monkeypatch) -> None:
    """Additional rate limits from payload are persisted via additional_usage_repo."""
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()
    monkeypatch.setattr("app.modules.usage.updater._now_epoch", lambda: 2000)

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "o-pro",
                        "metered_feature": "o_pro",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 50.0,
                                "reset_at": 1735689600,
                                "limit_window_seconds": 300,
                            },
                            "secondary_window": {
                                "used_percent": 75.0,
                                "reset_after_seconds": 120,
                                "limit_window_seconds": 3600,
                            },
                        },
                    }
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_add", "workspace_add", email="add@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    # Primary usage entry written
    assert len(usage_repo.entries) == 1

    # Two additional entries: primary + secondary window
    assert len(additional_repo.entries) == 2
    by_window = {e.window: e for e in additional_repo.entries}

    primary = by_window["primary"]
    assert primary.account_id == "acc_add"
    assert primary.limit_name == "o-pro"
    assert primary.metered_feature == "o_pro"
    assert primary.used_percent == 50.0
    assert primary.reset_at == 1735689600
    assert primary.window_minutes == 5

    secondary = by_window["secondary"]
    assert secondary.account_id == "acc_add"
    assert secondary.limit_name == "o-pro"
    assert secondary.metered_feature == "o_pro"
    assert secondary.used_percent == 75.0
    assert secondary.reset_at == 2120  # now_epoch(2000) + 120
    assert secondary.window_minutes == 60


@pytest.mark.asyncio
async def test_additional_rate_limits_normalize_known_alias_to_canonical_quota_key(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "GPT-5.3-Codex-Spark",
                        "metered_feature": "codex_bengalfox",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 12.0,
                                "reset_at": 1735689600,
                                "limit_window_seconds": 300,
                            }
                        },
                    }
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)

    await updater.refresh_accounts([_make_account("acc_alias", "workspace_alias")], latest_usage={})

    assert len(additional_repo.entries) == 1
    entry = additional_repo.entries[0]
    assert entry.quota_key == "codex_spark"
    assert entry.limit_name == "GPT-5.3-Codex-Spark"
    assert entry.metered_feature == "codex_bengalfox"


@pytest.mark.asyncio
async def test_additional_rate_limits_merge_aliases_before_pruning_quota(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "GPT-5.3-Codex-Spark",
                        "metered_feature": "codex_bengalfox",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 12.0,
                                "reset_at": 1735689600,
                                "limit_window_seconds": 300,
                            }
                        },
                    },
                    {
                        "limit_name": "codex_other",
                        "metered_feature": "codex_bengalfox",
                        "rate_limit": None,
                    },
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)

    await updater.refresh_accounts([_make_account("acc_alias_merge", "workspace_alias_merge")], latest_usage={})

    assert len(additional_repo.entries) == 1
    entry = additional_repo.entries[0]
    assert entry.quota_key == "codex_spark"
    assert entry.limit_name == "GPT-5.3-Codex-Spark"
    assert additional_repo.deleted_account_limit_pairs == []


@pytest.mark.asyncio
async def test_additional_rate_limits_merge_windows_across_aliases(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "GPT-5.3-Codex-Spark",
                        "metered_feature": "codex_bengalfox",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 12.0,
                                "reset_at": 1735689600,
                                "limit_window_seconds": 300,
                            }
                        },
                    },
                    {
                        "limit_name": "codex_other",
                        "metered_feature": "codex_bengalfox",
                        "rate_limit": {
                            "secondary_window": {
                                "used_percent": 33.0,
                                "reset_at": 1735689700,
                                "limit_window_seconds": 1800,
                            }
                        },
                    },
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)

    await updater.refresh_accounts([_make_account("acc_alias_windows", "workspace_alias_windows")], latest_usage={})

    assert len(additional_repo.entries) == 2
    by_window = {entry.window: entry for entry in additional_repo.entries}
    assert by_window["primary"].quota_key == "codex_spark"
    assert by_window["secondary"].quota_key == "codex_spark"
    assert additional_repo.deleted_account_limit_pairs == []


@pytest.mark.asyncio
async def test_additional_rate_limits_null_writes_nothing(monkeypatch) -> None:
    """When additional_rate_limits is null, no additional entries are written."""
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_null", "workspace_null", email="null@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(additional_repo.entries) == 0


@pytest.mark.asyncio
async def test_additional_rate_limits_sync_even_when_main_rate_limit_missing(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "additional_rate_limits": [
                    {
                        "limit_name": "o-pro",
                        "metered_feature": "o_pro",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 25.0,
                                "reset_at": 1735689600,
                                "limit_window_seconds": 60,
                            }
                        },
                    }
                ]
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_add_only", "workspace_add_only", email="add-only@example.com")

    refreshed = await updater.refresh_accounts([acc], latest_usage={})

    # Additional-only accounts write additional data and mark themselves as fresh
    # to prevent tight re-polling (R6-F1).
    assert refreshed is True
    assert usage_repo.entries == []
    assert len(additional_repo.entries) == 1
    assert additional_repo.entries[0].limit_name == "o-pro"


@pytest.mark.asyncio
async def test_additional_only_account_not_repolled_within_interval(monkeypatch) -> None:
    """R6-F1: Additional-only accounts must not cause tight re-polling."""
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    call_count = 0

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        nonlocal call_count
        call_count += 1
        return UsagePayload.model_validate(
            {
                "additional_rate_limits": [
                    {
                        "limit_name": "o-pro",
                        "metered_feature": "o_pro",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 25.0,
                                "reset_at": 1735689600,
                                "limit_window_seconds": 60,
                            }
                        },
                    }
                ]
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_add_only2", "workspace_add_only2", email="add-only2@example.com")

    # First call fetches usage.
    await updater.refresh_accounts([acc], latest_usage={})
    assert call_count == 1

    # Second call immediately should be skipped due to freshness cache.
    await updater.refresh_accounts([acc], latest_usage={})
    assert call_count == 1


@pytest.mark.asyncio
async def test_additional_rate_limits_empty_list_writes_nothing(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_empty", "workspace_empty", email="empty@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(additional_repo.entries) == 0
    assert additional_repo.deleted_account_ids == ["acc_empty"]


@pytest.mark.asyncio
async def test_additional_rate_limits_none_does_not_prune_existing_rows(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_none_preserve", "workspace_none_preserve", email="preserve@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert additional_repo.deleted_account_ids == []


@pytest.mark.asyncio
async def test_additional_rate_limits_multiple_limits(monkeypatch) -> None:
    """Multiple additional limits produce one entry per limit per window."""
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()
    monkeypatch.setattr("app.modules.usage.updater._now_epoch", lambda: 5000)

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 5.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "o-pro",
                        "metered_feature": "o_pro",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 30.0,
                                "reset_at": 9999,
                                "limit_window_seconds": 600,
                            },
                        },
                    },
                    {
                        "limit_name": "deep-research",
                        "metered_feature": "deep_research",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 80.0,
                                "reset_at": 8888,
                                "limit_window_seconds": 120,
                            },
                            "secondary_window": {
                                "used_percent": 40.0,
                                "reset_at": 7777,
                                "limit_window_seconds": 1800,
                            },
                        },
                    },
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_multi", "workspace_multi", email="multi@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    # o-pro: 1 primary; deep-research: 1 primary + 1 secondary = 3 total
    assert len(additional_repo.entries) == 3
    names = [(e.limit_name, e.window) for e in additional_repo.entries]
    assert ("o-pro", "primary") in names
    assert ("deep-research", "primary") in names
    assert ("deep-research", "secondary") in names


@pytest.mark.asyncio
async def test_additional_rate_limits_secondary_none_only_primary(monkeypatch) -> None:
    """When secondary_window is None, only primary entry is written."""
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "sora",
                        "metered_feature": "sora_video",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 60.0,
                                "reset_at": 4444,
                                "limit_window_seconds": 180,
                            },
                        },
                    }
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_sec_none", "workspace_sec_none", email="sec-none@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert len(additional_repo.entries) == 1
    entry = additional_repo.entries[0]
    assert entry.window == "primary"
    assert entry.limit_name == "sora"
    assert entry.metered_feature == "sora_video"
    assert entry.used_percent == 60.0


@pytest.mark.asyncio
async def test_additional_rate_limits_prune_stale_limit_names(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "o-pro",
                        "metered_feature": "o_pro",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 25.0,
                                "reset_at": 5555,
                                "limit_window_seconds": 60,
                            },
                        },
                    }
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    additional_repo.entries.extend(
        [
            AdditionalUsageEntry(
                account_id="acc_prune",
                limit_name="o-pro",
                metered_feature="o_pro",
                window="primary",
                used_percent=10.0,
                reset_at=1111,
                window_minutes=1,
            ),
            AdditionalUsageEntry(
                account_id="acc_prune",
                limit_name="legacy-limit",
                metered_feature="legacy_feature",
                window="primary",
                used_percent=90.0,
                reset_at=2222,
                window_minutes=5,
            ),
        ]
    )
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_prune", "workspace_prune", email="prune@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert additional_repo.deleted_account_limit_pairs == [("acc_prune", "legacy_limit")]


@pytest.mark.asyncio
async def test_additional_rate_limits_prune_stale_secondary_window(monkeypatch) -> None:
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "additional_rate_limits": [
                    {
                        "limit_name": "o-pro",
                        "metered_feature": "o_pro",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 25.0,
                                "reset_at": 5555,
                                "limit_window_seconds": 60,
                            },
                        },
                    }
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    additional_repo.entries.extend(
        [
            AdditionalUsageEntry(
                account_id="acc_secondary_prune",
                limit_name="o-pro",
                metered_feature="o_pro",
                window="primary",
                used_percent=10.0,
                reset_at=1111,
                window_minutes=1,
            ),
            AdditionalUsageEntry(
                account_id="acc_secondary_prune",
                limit_name="o-pro",
                metered_feature="o_pro",
                window="secondary",
                used_percent=80.0,
                reset_at=2222,
                window_minutes=60,
            ),
        ]
    )
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_secondary_prune", "workspace_secondary_prune", email="secondary-prune@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    assert additional_repo.deleted_account_limit_pairs == []
    assert additional_repo.deleted_account_limit_windows == [("acc_secondary_prune", "o_pro", "secondary")]


@pytest.mark.asyncio
async def test_additional_rate_limits_no_credits_passed(monkeypatch) -> None:
    """Credits data is NOT passed to additional limit entries (no credits_* fields)."""
    monkeypatch.setenv("CODEX_LB_USAGE_REFRESH_ENABLED", "true")
    from app.core.config.settings import get_settings

    get_settings.cache_clear()

    async def stub_fetch_usage(**_: Any) -> UsagePayload:
        return UsagePayload.model_validate(
            {
                "rate_limit": {
                    "primary_window": {
                        "used_percent": 10.0,
                        "reset_at": 1735689600,
                        "limit_window_seconds": 60,
                    },
                },
                "credits": {"has_credits": True, "unlimited": False, "balance": "100.0"},
                "additional_rate_limits": [
                    {
                        "limit_name": "o-pro",
                        "metered_feature": "o_pro",
                        "rate_limit": {
                            "primary_window": {
                                "used_percent": 25.0,
                                "reset_at": 5555,
                                "limit_window_seconds": 60,
                            },
                        },
                    }
                ],
            }
        )

    monkeypatch.setattr("app.modules.usage.updater.fetch_usage", stub_fetch_usage)

    usage_repo = StubUsageRepository()
    additional_repo = StubAdditionalUsageRepository()
    updater = UsageUpdater(usage_repo, accounts_repo=None, additional_usage_repo=additional_repo)
    acc = _make_account("acc_no_cred", "workspace_no_cred", email="no-cred@example.com")

    await updater.refresh_accounts([acc], latest_usage={})

    # Primary usage entry should have credits
    assert len(usage_repo.entries) == 1
    assert usage_repo.entries[0].credits_has is True

    # Additional entry should NOT have credits fields (not part of the protocol)
    assert len(additional_repo.entries) == 1
    entry = additional_repo.entries[0]
    assert not hasattr(entry, "credits_has")
    assert not hasattr(entry, "credits_unlimited")
    assert not hasattr(entry, "credits_balance")



"""fix postgresql enum value casing from uppercase to lowercase

Revision ID: 20260310_000000_fix_postgresql_enum_value_casing
Revises: 20260309_000000_add_additional_usage_history
Create Date: 2026-03-10
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "20260310_000000_fix_postgresql_enum_value_casing"
down_revision = "20260309_000000_add_additional_usage_history"
branch_labels = None
depends_on = None

_ENUM_RENAMES: dict[str, list[tuple[str, str]]] = {
    "account_status": [
        ("ACTIVE", "active"),
        ("RATE_LIMITED", "rate_limited"),
        ("QUOTA_EXCEEDED", "quota_exceeded"),
        ("PAUSED", "paused"),
        ("DEACTIVATED", "deactivated"),
    ],
    "limit_type": [
        ("TOTAL_TOKENS", "total_tokens"),
        ("INPUT_TOKENS", "input_tokens"),
        ("OUTPUT_TOKENS", "output_tokens"),
        ("COST_USD", "cost_usd"),
    ],
    "limit_window": [
        ("DAILY", "daily"),
        ("WEEKLY", "weekly"),
        ("MONTHLY", "monthly"),
    ],
}


def _enum_value_exists(bind: sa.engine.Connection, enum_type_name: str, enum_value: str) -> bool:
    result = bind.execute(
        sa.text(
            "SELECT 1 FROM pg_enum e "
            "JOIN pg_type t ON e.enumtypid = t.oid "
            "WHERE t.typname = :type_name AND e.enumlabel = :value"
        ),
        {"type_name": enum_type_name, "value": enum_value},
    )
    return result.scalar() is not None


def _rename_enum_values(
    bind: sa.engine.Connection,
    enum_type_name: str,
    renames: list[tuple[str, str]],
) -> None:
    for old_value, new_value in renames:
        if _enum_value_exists(bind, enum_type_name, old_value):
            bind.execute(sa.text(f"ALTER TYPE {enum_type_name} RENAME VALUE '{old_value}' TO '{new_value}'"))


def upgrade() -> None:
    bind = op.get_bind()
    if bind.dialect.name != "postgresql":
        return

    for enum_type_name, renames in _ENUM_RENAMES.items():
        _rename_enum_values(bind, enum_type_name, renames)


def downgrade() -> None:
    bind = op.get_bind()
    if bind.dialect.name != "postgresql":
        return

    for enum_type_name, renames in _ENUM_RENAMES.items():
        reverse_renames = [(new_val, old_val) for old_val, new_val in renames]
        _rename_enum_values(bind, enum_type_name, reverse_renames)



from __future__ import annotations

import hashlib
from dataclasses import dataclass
from datetime import datetime

from sqlalchemy import delete, func, select, text, update
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import AsyncSession

from app.db.models import Account, AccountStatus, DashboardSettings, RequestLog, StickySession, UsageHistory

_SETTINGS_ROW_ID = 1
_DUPLICATE_ACCOUNT_SUFFIX = "__copy"


@dataclass(frozen=True, slots=True)
class AccountRequestUsageSummary:
    request_count: int
    total_tokens: int
    cached_input_tokens: int
    total_cost_usd: float


class AccountIdentityConflictError(Exception):
    def __init__(self, email: str) -> None:
        self.email = email
        super().__init__(
            f"Cannot overwrite account for email '{email}' because multiple matching accounts exist. "
            "Remove duplicates or enable import without overwrite."
        )


class AccountsRepository:
    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def get_by_id(self, account_id: str) -> Account | None:
        return await self._session.get(Account, account_id)

    async def list_accounts(self) -> list[Account]:
        result = await self._session.execute(select(Account).order_by(Account.email))
        return list(result.scalars().all())

    async def list_request_usage_summary_by_account(
        self,
        account_ids: list[str] | None = None,
    ) -> dict[str, AccountRequestUsageSummary]:
        return await self.list_request_usage_summary_by_subject(account_ids)

    async def list_request_usage_summary_by_subject(
        self,
        routing_subject_ids: list[str] | None = None,
    ) -> dict[str, AccountRequestUsageSummary]:
        summaries: dict[str, AccountRequestUsageSummary] = {}
        subject_id = func.coalesce(RequestLog.routing_subject_id, RequestLog.account_id)
        output_tokens_expr = func.coalesce(RequestLog.output_tokens, RequestLog.reasoning_tokens, 0)
        stmt = select(
            subject_id.label("routing_subject_id"),
            func.count(RequestLog.id).label("request_count"),
            func.coalesce(func.sum(RequestLog.input_tokens), 0).label("input_tokens"),
            func.coalesce(func.sum(output_tokens_expr), 0).label("output_tokens"),
            func.coalesce(func.sum(RequestLog.cached_input_tokens), 0).label("cached_input_tokens"),
            func.coalesce(func.sum(RequestLog.cost_usd), 0.0).label("total_cost_usd"),
        ).group_by(subject_id)
        if routing_subject_ids:
            stmt = stmt.where(subject_id.in_(routing_subject_ids))

        result = await self._session.execute(stmt)
        for (
            routing_subject_id,
            request_count,
            input_tokens,
            output_tokens,
            cached_input_tokens,
            total_cost_usd,
        ) in result.all():
            if not routing_subject_id:
                continue
            input_sum = int(input_tokens or 0)
            output_sum = int(output_tokens or 0)
            cached_sum = int(cached_input_tokens or 0)
            cached_sum = max(0, min(cached_sum, input_sum))
            return_row = AccountRequestUsageSummary(
                request_count=int(request_count or 0),
                total_tokens=input_sum + output_sum,
                cached_input_tokens=cached_sum,
                total_cost_usd=round(float(total_cost_usd or 0.0), 6),
            )
            summaries[routing_subject_id] = return_row

        return summaries

    async def exists_active_chatgpt_account_id(self, chatgpt_account_id: str) -> bool:
        result = await self._session.execute(
            select(Account.id)
            .where(Account.chatgpt_account_id == chatgpt_account_id)
            .where(Account.status.notin_((AccountStatus.DEACTIVATED, AccountStatus.PAUSED)))
            .limit(1)
        )
        return result.scalar_one_or_none() is not None

    async def upsert(self, account: Account, *, merge_by_email: bool | None = None) -> Account:
        dialect_name = self._dialect_name()
        sqlite_lock_acquired = False
        if merge_by_email is None:
            if dialect_name == "sqlite":
                await self._acquire_sqlite_merge_lock()
                sqlite_lock_acquired = True
            merge_by_email = await self._merge_by_email_enabled()

        if merge_by_email:
            if dialect_name == "sqlite" and not sqlite_lock_acquired:
                await self._acquire_sqlite_merge_lock()
            elif dialect_name == "postgresql":
                await self._acquire_postgresql_merge_lock(account.email)
        else:
            if dialect_name == "sqlite" and not sqlite_lock_acquired:
                await self._acquire_sqlite_merge_lock()
            elif dialect_name == "postgresql":
                await self._acquire_postgresql_identity_lock(account.id)

        existing = await self._session.get(Account, account.id)
        if existing:
            if merge_by_email:
                _apply_account_updates(existing, account)
                await self._session.commit()
                await self._session.refresh(existing)
                return existing
            account.id = await self._next_available_account_id(account.id)

        if merge_by_email:
            existing_by_email = await self._single_account_by_email(account.email)
            if existing_by_email:
                _apply_account_updates(existing_by_email, account)
                await self._session.commit()
                await self._session.refresh(existing_by_email)
                return existing_by_email

        self._session.add(account)
        await self._session.commit()
        await self._session.refresh(account)
        return account

    async def update_status(
        self,
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
        reset_at: int | None = None,
    ) -> bool:
        result = await self._session.execute(
            update(Account)
            .where(Account.id == account_id)
            .values(status=status, deactivation_reason=deactivation_reason, reset_at=reset_at)
            .returning(Account.id)
        )
        await self._session.commit()
        return result.scalar_one_or_none() is not None

    async def update_status_if_current(
        self,
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
        reset_at: int | None = None,
        *,
        expected_status: AccountStatus,
        expected_deactivation_reason: str | None = None,
        expected_reset_at: int | None = None,
    ) -> bool:
        stmt = (
            update(Account)
            .where(Account.id == account_id)
            .where(Account.status == expected_status)
            .values(status=status, deactivation_reason=deactivation_reason, reset_at=reset_at)
            .returning(Account.id)
        )
        if expected_deactivation_reason is None:
            stmt = stmt.where(Account.deactivation_reason.is_(None))
        else:
            stmt = stmt.where(Account.deactivation_reason == expected_deactivation_reason)
        if expected_reset_at is None:
            stmt = stmt.where(Account.reset_at.is_(None))
        else:
            stmt = stmt.where(Account.reset_at == expected_reset_at)
        result = await self._session.execute(stmt)
        await self._session.commit()
        return result.scalar_one_or_none() is not None

    async def delete(self, account_id: str) -> bool:
        await self._session.execute(delete(UsageHistory).where(UsageHistory.account_id == account_id))
        await self._session.execute(delete(RequestLog).where(RequestLog.account_id == account_id))
        await self._session.execute(delete(StickySession).where(StickySession.account_id == account_id))
        result = await self._session.execute(delete(Account).where(Account.id == account_id).returning(Account.id))
        await self._session.commit()
        return result.scalar_one_or_none() is not None

    async def update_tokens(
        self,
        account_id: str,
        access_token_encrypted: bytes,
        refresh_token_encrypted: bytes,
        id_token_encrypted: bytes,
        last_refresh: datetime,
        plan_type: str | None = None,
        email: str | None = None,
        chatgpt_account_id: str | None = None,
    ) -> bool:
        values: dict[str, bytes | datetime | str] = {
            "access_token_encrypted": access_token_encrypted,
            "refresh_token_encrypted": refresh_token_encrypted,
            "id_token_encrypted": id_token_encrypted,
            "last_refresh": last_refresh,
        }
        if plan_type is not None:
            values["plan_type"] = plan_type
        if email is not None:
            values["email"] = email
        if chatgpt_account_id is not None:
            values["chatgpt_account_id"] = chatgpt_account_id
        result = await self._session.execute(
            update(Account).where(Account.id == account_id).values(**values).returning(Account.id)
        )
        await self._session.commit()
        return result.scalar_one_or_none() is not None

    async def _merge_by_email_enabled(self) -> bool:
        settings = await self._session.get(DashboardSettings, _SETTINGS_ROW_ID)
        if settings is None:
            return True
        return not settings.import_without_overwrite

    async def _next_available_account_id(self, base_id: str) -> str:
        candidate = base_id
        sequence = 2
        while await self._session.get(Account, candidate) is not None:
            candidate = f"{base_id}{_DUPLICATE_ACCOUNT_SUFFIX}{sequence}"
            sequence += 1
        return candidate

    async def _single_account_by_email(self, email: str) -> Account | None:
        result = await self._session.execute(
            select(Account).where(Account.email == email).order_by(Account.created_at.asc(), Account.id.asc()).limit(2)
        )
        matches = list(result.scalars().all())
        if not matches:
            return None
        if len(matches) > 1:
            raise AccountIdentityConflictError(email)
        return matches[0]

    def _dialect_name(self) -> str:
        return self._session.get_bind().dialect.name

    async def _acquire_sqlite_merge_lock(self) -> None:
        try:
            await self._session.execute(text("BEGIN IMMEDIATE"))
        except OperationalError as exc:
            message = str(exc).lower()
            if "within a transaction" not in message:
                raise
            # A no-op write escalates the current deferred transaction to a write
            # transaction, serializing concurrent writers.
            await self._session.execute(text("UPDATE accounts SET id = id WHERE 1 = 0"))

    async def _acquire_postgresql_merge_lock(self, email: str) -> None:
        lock_key = _advisory_lock_key("merge-email", email)
        await self._session.execute(
            text("SELECT pg_advisory_xact_lock(:lock_key)"),
            {"lock_key": lock_key},
        )

    async def _acquire_postgresql_identity_lock(self, account_id: str) -> None:
        lock_key = _advisory_lock_key("account-id", account_id)
        await self._session.execute(
            text("SELECT pg_advisory_xact_lock(:lock_key)"),
            {"lock_key": lock_key},
        )


def _apply_account_updates(target: Account, source: Account) -> None:
    target.chatgpt_account_id = source.chatgpt_account_id
    target.email = source.email
    target.plan_type = source.plan_type
    target.access_token_encrypted = source.access_token_encrypted
    target.refresh_token_encrypted = source.refresh_token_encrypted
    target.id_token_encrypted = source.id_token_encrypted
    target.last_refresh = source.last_refresh
    target.status = source.status
    target.deactivation_reason = source.deactivation_reason


def _advisory_lock_key(scope: str, value: str) -> int:
    digest = hashlib.sha256(f"{scope}:{value}".encode("utf-8")).digest()
    return int.from_bytes(digest[:8], byteorder="big", signed=True)



"""add api_key_limits table and migrate legacy weekly limits

Revision ID: 20260214_000000_add_api_key_limits
Revises: 20260213_000800_add_api_keys
Create Date: 2026-02-14
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260214_000000_add_api_key_limits"
down_revision = "20260213_000800_add_api_keys"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {column["name"] for column in inspector.get_columns(table_name)}


def upgrade() -> None:
    bind = op.get_bind()

    if not _table_exists(bind, "api_key_limits"):
        op.create_table(
            "api_key_limits",
            sa.Column("id", sa.Integer(), primary_key=True, autoincrement=True),
            sa.Column(
                "api_key_id",
                sa.String(),
                sa.ForeignKey("api_keys.id", ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column("limit_type", sa.String(), nullable=False),
            sa.Column("limit_window", sa.String(), nullable=False),
            sa.Column("max_value", sa.BigInteger(), nullable=False),
            sa.Column(
                "current_value",
                sa.BigInteger(),
                nullable=False,
                server_default=sa.text("0"),
            ),
            sa.Column("model_filter", sa.String(), nullable=True),
            sa.Column("reset_at", sa.DateTime(), nullable=False),
        )
        op.create_index(
            "idx_api_key_limits_key_id",
            "api_key_limits",
            ["api_key_id"],
            unique=False,
        )

    # Migrate existing weekly_token_limit data to api_key_limits rows
    api_keys_cols = _columns(bind, "api_keys")
    if "weekly_token_limit" in api_keys_cols:
        api_keys = sa.table(
            "api_keys",
            sa.column("id", sa.String),
            sa.column("weekly_token_limit", sa.Integer),
            sa.column("weekly_tokens_used", sa.Integer),
            sa.column("weekly_reset_at", sa.DateTime),
        )
        api_key_limits = sa.table(
            "api_key_limits",
            sa.column("api_key_id", sa.String),
            sa.column("limit_type", sa.String),
            sa.column("limit_window", sa.String),
            sa.column("max_value", sa.BigInteger),
            sa.column("current_value", sa.BigInteger),
            sa.column("model_filter", sa.String),
            sa.column("reset_at", sa.DateTime),
        )

        result = bind.execute(
            sa.select(
                api_keys.c.id,
                api_keys.c.weekly_token_limit,
                api_keys.c.weekly_tokens_used,
                api_keys.c.weekly_reset_at,
            ).where(api_keys.c.weekly_token_limit.isnot(None))
        )

        for row in result:
            bind.execute(
                api_key_limits.insert().values(
                    api_key_id=row.id,
                    limit_type="total_tokens",
                    limit_window="weekly",
                    max_value=row.weekly_token_limit,
                    current_value=row.weekly_tokens_used,
                    model_filter=None,
                    reset_at=row.weekly_reset_at,
                )
            )

        # Drop legacy columns (batch for SQLite compatibility)
        with op.batch_alter_table("api_keys") as batch_op:
            batch_op.drop_column("weekly_token_limit")
            batch_op.drop_column("weekly_tokens_used")
            batch_op.drop_column("weekly_reset_at")


def downgrade() -> None:
    bind = op.get_bind()

    # Re-add legacy columns
    api_keys_cols = _columns(bind, "api_keys")
    with op.batch_alter_table("api_keys") as batch_op:
        if "weekly_token_limit" not in api_keys_cols:
            batch_op.add_column(sa.Column("weekly_token_limit", sa.Integer(), nullable=True))
        if "weekly_tokens_used" not in api_keys_cols:
            batch_op.add_column(
                sa.Column(
                    "weekly_tokens_used",
                    sa.Integer(),
                    nullable=False,
                    server_default=sa.text("0"),
                )
            )
        if "weekly_reset_at" not in api_keys_cols:
            batch_op.add_column(
                sa.Column(
                    "weekly_reset_at",
                    sa.DateTime(),
                    nullable=True,
                )
            )

    if _table_exists(bind, "api_key_limits"):
        op.drop_index("idx_api_key_limits_key_id", table_name="api_key_limits")
        op.drop_table("api_key_limits")



from __future__ import annotations

import pytest

from app.core.resilience import retry_budget as retry_budget_module
from app.core.resilience.retry_budget import RetryBudget

pytestmark = pytest.mark.unit


@pytest.mark.asyncio
async def test_retry_budget_allows_retries_up_to_limit():
    budget = RetryBudget(max_retries_per_window=3, window_seconds=60)

    assert await budget.try_acquire() is True
    assert await budget.try_acquire() is True
    assert await budget.try_acquire() is True
    assert budget.remaining() == 0


@pytest.mark.asyncio
async def test_retry_budget_rejects_after_limit():
    budget = RetryBudget(max_retries_per_window=2, window_seconds=60)

    assert await budget.try_acquire() is True
    assert await budget.try_acquire() is True
    assert await budget.try_acquire() is False


@pytest.mark.asyncio
async def test_retry_budget_allows_new_retry_after_window_expires(monkeypatch: pytest.MonkeyPatch):
    values = [100.0, 101.0, 107.0, 107.0]
    state = {"index": 0}

    def fake_monotonic() -> float:
        index = state["index"]
        if index < len(values):
            state["index"] = index + 1
            return values[index]
        return values[-1]

    monkeypatch.setattr(retry_budget_module.time, "monotonic", fake_monotonic)

    budget = RetryBudget(max_retries_per_window=2, window_seconds=5)

    assert await budget.try_acquire() is True
    assert await budget.try_acquire() is True
    assert await budget.try_acquire() is True
    assert budget.remaining() == 1



from __future__ import annotations


class ClientPayloadError(ValueError):
    def __init__(self, message: str, *, param: str | None = None) -> None:
        super().__init__(message)
        self.param = param
